/* vi: set ts=4 sw=4: */
#include <string.h>
#include <fcntl.h>
#include "names.h"
//#include "physical.h"
//#include "db_mod.h"
#include "iface.h"

#include "pkg_set.h"
#include "installed_db_manager.h"

using namespace database;

int db_manager::null_callback(void *pArg, int argc, char **argv, char **columnNames) {
	return 0;
}

int db_manager::count_callback(void *pArg, int argc, char **argv, char **columnNames) {
	(*((int *)pArg))++;
	return 0;
}

int db_manager::search_callback(void *pArg, int argc, char **argv, char **columnNames) {
	int i;

	for(i=0; i<argc; i++)
		printf("%s: %s\n", columnNames[i], argv[i]);
	printf("\n");
	return 0;
}

int db_manager::search_from_name_callback(void *pArg, int argc, char **argv, char **columnNames) {
	Pkg_set *result = (Pkg_set*)pArg;
	if(argc == 2) {
		Package p(argv[0], argv[1]);
		result->add(p);
	}
	return 0;
}


static sqlite *_installed_sqlite_db=NULL;
static sqlite *_available_sqlite_db=NULL;

sqlite *database::get_installed() {return _installed_sqlite_db;}
sqlite *database::get_available() {return _available_sqlite_db;}

bool db_manager::db_init(abpmString &destdir) {
	database::db_init(destdir);
	switch(db_type) {
		case INSTALLED:
			mydb = _installed_sqlite_db;
			break;
		case AVAILABLE:
			mydb = _available_sqlite_db;
			break;
	}
#ifndef DUMMY
/*	char *error;
	if(mydb==NULL) {
		abpmString db_filename;
		switch(type) {
			case INSTALLED:
				db_filename=destdir+ABPM_INSTALLED_DB_FILE;
				break;
			case AVAILABLE:
				db_filename=destdir+ABPM_AVAILABLE_DB_FILE;
				break;
		}
		mydb = sqlite_open(db_filename, O_RDWR, &error);
		if(mydb==NULL) {
			fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
			//fprintf(stderr, "%s\n", error);
			return false;
		} //else
		//	delete[] error;
		
		sqlite_exec_printf(mydb, "pragma default_cache_size=40000;", NULL, NULL, NULL);
		sqlite_exec_printf(mydb, "pragma cache_size=40000;", NULL, NULL, NULL);
		sqlite_exec_printf(mydb, "pragma default_synchronous=off;", NULL, NULL, NULL);
		sqlite_exec_printf(mydb, "pragma synchronous=off;", NULL, NULL, NULL);
	}*/
#endif
	return true;
}

bool db_manager::db_uninit() {
#ifndef DUMMY
//	sqlite_close(mydb);
#endif
	return true;
}

bool db_manager::begin_transaction() {
	char *error;

	if(sqlite_exec(mydb, "BEGIN TRANSACTION", installed_db_manager::null_callback, NULL, &error))
		fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
		//fprintf(stderr, "%s\n", error);
	
	// Elvileg ha ez elszall, akkor meg lehet jo...
	return true;
}


char *db_manager::generate_metapackage_name(char **item) {
	char *name;
	int len;

	len = strlen(item[PKG_NAME]) + strlen(item[PKG_VERSION]) + 70;
	name = new char[len];
	
	sprintf(name, "#%s#%s#%d", item[PKG_NAME], item[PKG_VERSION], utolso);
	utolso++;
	
	return name;
}

bool db_manager::write_deps_to_db(int OR_enable, char **item, Entry_type t, char *command) {
	/*                                                                        ^^^^^^^^^^^^^
	 * This variable is only a buffer. I just get it from caller function
	 * because I don't want to allocate memory on every call of this function*/

	char *error;
	char *tmp1, *tmp2;
	char *metapackage_name=NULL;
	//char *command;
	int len;
	int i;
	int pos = 0;
	int need_metapackage;
	
	len = strlen(item[t]) + 1;
	tmp1 = new char[len];
	tmp2 = new char[len];
	
	while(item[t][pos] != '\0') {
		need_metapackage = 0;
		while(item[t][pos] !=',' && item[t][pos] != '\0') {
			tmp2[0] = '\0';

			for(i=pos; item[t][i]!=',' &&
					item[t][i] != '\0' &&
					item[t][i] != '(' &&
					item[t][i] != '|' &&
					item[t][i] != ' '; i++)
				tmp1[i-pos] = item[t][i];
			tmp1[i-pos] = '\0';
			
			while(item[t][i] == ' ' || item[t][i] == '\t') i++;
			
			pos = i;
			
			if(item[t][i] == '(') {
				pos = i + 1;
					for(i=pos; item[t][i]!=')' &&
							item[t][i] != '\0'; i++)
						tmp2[i - pos] = item[t][i];
					tmp2[i - pos] = '\0';
					pos = i;
					if(item[t][pos] != ')') {
						fprintf(stderr, "parse error in line:"
								"\n\t%s %s\n",
								Entry_name[i], item[t]);
						return false;
					}
					pos++;
			}
			while(item[t][pos] == ' ' || item[t][pos] == '\t') pos++;
			if(item[t][pos] == '|') {
				if(OR_enable) { 
					if(!need_metapackage) {
						need_metapackage = 1;
						metapackage_name = generate_metapackage_name(item);
					}
					pos++;
					while(item[t][pos] == ' ' || item[t][pos] == '\t') pos++;
				} else {
					fprintf(stderr, "parse error: '|' char is not allowed"
							"in %s line in package %s\n",
							Entry_name[i], item[PKG_NAME]);
					return false;
				}
			}
			if(need_metapackage) {
				sprintf(command, "INSERT INTO provides"
						"(name, version, dname, dversion)"
						"VALUES ('%s','%s','%s','%s')",
						tmp1, tmp2, 
						metapackage_name, "");
//#ifdef DEBUG
//				fprintf(stderr, "%s:%d:\t command = %s\n",__FILE__,__LINE__, command);
//#endif
#ifndef DUMMY
				if(sqlite_exec(mydb, command, installed_db_manager::null_callback, NULL, &error)) {
					fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
					//fprintf(stderr, "%s\n", error);
					fprintf(stderr, "command:%s\n", command);
//					fprintf(stderr, "Press ENTER to continue...\n");
//					fgetc(stdin); //debug
				}
#endif
			}
		} /* ',' */
		if(item[t][pos] != '\0')
			pos++;
		
		if(need_metapackage)
			tmp2[0]='\0';
		else
			metapackage_name = tmp1;
		
		sprintf(command, "INSERT INTO %s (name, version, dname, dversion)"
				"VALUES ('%s','%s','%s','%s')",
				Table_name[t],
				item[PKG_NAME],
				item[PKG_VERSION],
				metapackage_name,
				tmp2);
#ifdef DEBUG		
		printf("pos = %d\n", pos);
		printf("\t command = %s\n", command);
#endif
#ifndef DUMMY
		if(sqlite_exec(mydb, command, installed_db_manager::null_callback, NULL, &error)) {
			//fprintf(stderr, "%s\n", error);
			fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
			fprintf(stderr, "command:%s\n", command);
			//fprintf(stderr, "Press ENTER to continue...\n");
			//fgetc(stdin);
		}
#endif	
		//sprintf(command, "COMMIT");
		/*if(sqlite_exec(mydb, command, installed_db_manager::null_callback, NULL, &error)) {
			fprintf(stderr, "%s\n", error);
			fprintf(stderr, "command:%s\n", command);
			fprintf(stderr, "Press ENTER to continue...\n");
			fgetc(stdin);
		}*/
		
		if(need_metapackage)
			delete[] metapackage_name;
		
	}
	delete[] tmp1;
	delete[] tmp2;
	return true;
}

bool database::db_init(abpmString &destdir) {
	char *error;
	abpmString db_filename=destdir+ABPM_INSTALLED_DB_FILE;
	if(_installed_sqlite_db == NULL) {
		_installed_sqlite_db = sqlite_open(db_filename, O_RDWR, &error);
		if(_installed_sqlite_db==NULL) {
			fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
			//fprintf(stderr, "%s\n", error);
			return false;
		} /*else
			delete[] error;*/
		
		sqlite_exec_printf(_installed_sqlite_db, "pragma default_cache_size=40000;", NULL, NULL, NULL);
		sqlite_exec_printf(_installed_sqlite_db, "pragma cache_size=40000;", NULL, NULL, NULL);
		sqlite_exec_printf(_installed_sqlite_db, "pragma default_synchronous=off;", NULL, NULL, NULL);
		sqlite_exec_printf(_installed_sqlite_db, "pragma synchronous=off;", NULL, NULL, NULL);
	}
	db_filename=destdir+ABPM_AVAILABLE_DB_FILE;
	if(_available_sqlite_db == NULL) {
		_available_sqlite_db = sqlite_open(db_filename, O_RDWR, &error);
		if(_available_sqlite_db==NULL) {
			fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
			//fprintf(stderr, "%s\n", error);
			return false;
		} /*else
			delete[] error;*/
		
		sqlite_exec_printf(_available_sqlite_db, "pragma default_cache_size=40000;", NULL, NULL, NULL);
		sqlite_exec_printf(_available_sqlite_db, "pragma cache_size=40000;", NULL, NULL, NULL);
		sqlite_exec_printf(_available_sqlite_db, "pragma default_synchronous=off;", NULL, NULL, NULL);
		sqlite_exec_printf(_available_sqlite_db, "pragma synchronous=off;", NULL, NULL, NULL);
	}
	return true;
}

bool database::db_close() {
	sqlite_close(_installed_sqlite_db);
	_installed_sqlite_db=NULL;
	sqlite_close(_available_sqlite_db);
	_available_sqlite_db=NULL;
	return true;
}
	

/*
bool database::create_installed_db(Create_mode mode) {
	installed_db_manager dbgen;
	
	dbgen.db_init();
	dbgen.create_tables();

	if(mode != EMPTY) {
		dbgen.begin_transaction();
		dbgen.read_file();
		dbgen.end_transaction();
	}
	dbgen.db_uninit();
	
	return true;
}*/
