/* vi: set ts=4 sw=4: */
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>
#include "names.h"
#include "package.h"
#include "iface.h"
#include "abpmString.h"

#include "installed_db_manager.h"

using namespace database;

bool installed_db_manager::commit() {
	char *error;
	if(sqlite_exec(mydb, "COMMIT", installed_db_manager::null_callback, NULL, &error)) {
		fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
		//fprintf(stderr, "%s\n", error);
		return false;
	}
	return true;
}

bool installed_db_manager::end_transaction() {
	char *error;

//#ifdef GEN_DB
	char command[1024];
	int i;

	//if(sqlite_exec(mydb, "END TRANSACTION", installed_db_manager::null_callback, NULL, &error))
	
	for(i=0; i<PKG_NAME; i++) {
		sprintf(command, "CREATE INDEX index%d ON %s (name, version)",
				2 * i, Table_name[i]);
		if(sqlite_exec(mydb, command, installed_db_manager::null_callback, NULL, &error))
			fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
			//fprintf(stderr, "%s\n", error);
		sprintf(command, "CREATE INDEX index%d ON %s (dname, dversion)",
				2 * i + 1, Table_name[i]);
		if(sqlite_exec(mydb, command, installed_db_manager::null_callback, NULL, &error))
			fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
			//fprintf(stderr, "%s\n", error);
	}
	
	sprintf(command, "CREATE INDEX details_index ON details (name, version)");
	if(sqlite_exec(mydb, command, installed_db_manager::null_callback, NULL, &error))
		fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
		//fprintf(stderr, "%s\n", error);
	
	sprintf(command, "CREATE INDEX files_index1 ON files (name, version)");
	if(sqlite_exec(mydb, command, installed_db_manager::null_callback, NULL, &error))
		fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
		//fprintf(stderr, "%s\n", error);
	
	sprintf(command, "CREATE INDEX files_index2 ON files (filename)");
	if(sqlite_exec(mydb, command, installed_db_manager::null_callback, NULL, &error))
		fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
		//fprintf(stderr, "%s\n", error);
	
	
//#endif
	return commit();
}
	

bool installed_db_manager::create_tables() {
#ifndef DUMMY
	char *error;
	int i;
	char command[1024];

	sprintf(command, "CREATE TABLE details (name, version, section, priority, arch, size, maintainer, source, md5, short_desc, long_desc)");
	
	if(sqlite_exec(mydb, command, installed_db_manager::null_callback, NULL, &error)) {
		fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
		//fprintf(stderr, "%s\n", error);
		fprintf(stderr, "command: %s\n", command);
	}

	sprintf(command, "CREATE TABLE files (name, version, filename, file_owner, file_group, type, permission, size, md5, conffile)");
	if(sqlite_exec(mydb, command, installed_db_manager::null_callback, NULL, &error)) {
		fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
		//fprintf(stderr, "%s\n", error);
		fprintf(stderr, "command: %s\n", command);
//		fprintf(stderr, "Press Enter to continue...\n");
//		fgetc(stdin);
	}
	
	
	for(i=0; i<PKG_NAME; i++) {
		sprintf(command,"CREATE TABLE %s (name, version, dname, dversion)",Table_name[i]);
		if(sqlite_exec(mydb, command, installed_db_manager::null_callback, NULL, &error)) {
			fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
			//fprintf(stderr, "%s\n", error);
			fprintf(stderr, "command: %s\n", command);
		}
	}
#endif
	return true;
}


int item_length(char **item);

bool installed_db_manager::write_to_db(char **item, char *url_prefix/*unused*/) {
	char *command;
	char *error;
	int i;
	
	if(item==NULL)
		return false;
	
	command = new char[item_length(item) + 165];
				/*it's a litle more then I need...*/

	sprintf(command, "INSERT INTO details"
			"(name, version, section, priority, arch, size,"
				" maintainer, source, md5, short_desc, long_desc)"
			"VALUES ('%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s')",
			item[PKG_NAME],
			item[PKG_VERSION],
			item[PKG_SECTION],
			item[PKG_PRIORITY],
			item[PKG_ARCH],
			item[PKG_SIZE],
			item[PKG_MAINTAINER],
			item[PKG_SOURCE],
			item[PKG_MD5],
			item[PKG_SHORT_DESC],
			item[PKG_LONG_DESC]
			);
	
//	printf("command = %s\n", command);
#ifndef DUMMY
	if(sqlite_exec(mydb, command, installed_db_manager::null_callback, NULL, &error)) {
		fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
		//fprintf(stderr, "%s\n", error);
		fprintf(stderr, "command:%s\n", command);
//		fprintf(stderr, "Press ENTER to continue...\n");
//		fgetc(stdin);
	}
#endif
	//printf("sqlite_exec (((details)) megvolt\n");
	for(i=0; i<PKG_NAME; i++)
		if(strcmp(item[i], ""))
			write_deps_to_db(1, item, (Entry_type)i, command);
	/*TODO: file lists and md5 lists from /var/lib/dpkg/info/... */
	delete[] command;
	return true;
}


bool installed_db_manager::read_packages_file() {
//	FILE *packages;
	//char **result;
	//int nrow, ncolumn;
//	char *data;
	int pkg_num;
	//char **item;
	package_control_info *info;
	int i;
//	int result;
//	int end_section = 1;
	
//	file_reader *fr = new normal_file_reader(STATUS_FILENAME);

//	data = new char[1024];
	
	abpmString status_filename((char *)STATUS_FILENAME);
	physical::read_old_available_db(status_filename,  &info, &pkg_num);
	for (i=0; i<pkg_num; i++) {
		write_to_db(info[i].items, NULL);
		write_stat(info[i].items);
	}
//			if(end_section) {
//				write_to_db(item);
//				write_stat(item);
//			}
//	md5_close_and_free();
//	delete fr;
	return true;
}


bool installed_db_manager::write_physicalinfo_to_db(package_control_info *c_info, physical_package_info *ph_info) {
	char *command;
	int command_len;
	char *error;
	int i;
	char type_str[2];
	char mode_str[100];
	char size_str[100];
	
	//_debuginfo(c_info->items[PKG_NAME]);
	//fprintf(stderr, "\n");
	for(i=0; i<ph_info->size(); i++) {
		char *pkg_name = c_info->items[PKG_NAME]; 
		char *pkg_version = c_info->items[PKG_VERSION];
		abpmString filename = ph_info->get(i).filename;
		char *owner = ph_info->get(i).owner == NULL ? (char *)"" : ph_info->get(i).owner;
		char *group = ph_info->get(i).group == NULL ? (char *)"" : ph_info->get(i).group;
		char *type;
		if(ph_info->get(i).type_used) {
			sprintf(type_str, "%c", ph_info->get(i).type);
			type = type_str;
		} else {
			type = (char *)"";
		}
		
		char *mode;
		if(ph_info->get(i).mode_used) {
			sprintf(mode_str, "0%o", ph_info->get(i).mode);
			mode = mode_str;
		}else {
			mode = (char *)"";
		}
		char *md5 = ph_info->get(i).md5 == NULL ? (char *)"" : ph_info->get(i).md5;
		char *size;
		if(ph_info->get(i).size_used) {
			sprintf(size_str, "%d", ph_info->get(i).size);
			size = size_str;
		} else {
			size = (char *)"";
		}
		
		command_len =
			strlen(pkg_name) + 
			strlen(pkg_version) + 
			filename.getLength() +
			strlen(owner) + 
			strlen(group) + 
			strlen(type) +
			strlen(mode) +
			strlen(md5) +
			strlen(size) + 
			200;

		command = new char[command_len];
		/*
		 int count = 0;
		 sprintf(command, "SELECT name, version FROM files WHERE name='%s' AND version='%s'");
		 if(sqlite_exec(mydb, command, installed_db_manager::count_callback, &count, &error)) {
		 	fprintf(stderr, "%s\n", error);
			fprintf(stderr, "command: %s\n", command);
		}
		if(count == 0) {
		*/
		sprintf(command, "INSERT INTO files "
				"(name, version, filename, file_owner, file_group, "
				"type, permission, md5, size) "
				"VALUES ('%s','%s','%s','%s','%s','%s','%s','%s','%s')",
				pkg_name,
				pkg_version,
				filename.getstring(),
				owner,
				group,
				type,
				mode,
				md5,
				size);
		if(sqlite_exec(mydb, command, installed_db_manager::null_callback, NULL, &error)) {
			//fprintf(stderr, "%s\n", error);
			fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
			fprintf(stderr, "command:%s\n", command);
//			fprintf(stderr, "Press ENTER to continue...\n");
//			fgetc(stdin); //debug
		}
		/*
		} else {
			error/warning message?
			alter table?
		}
		*/
		delete[] command;
	}
	return true;
}

char *mystringcpy(char * str) {
	char *retval = new char[strlen(str) + 1];
	strcpy(retval, str);
	return retval;
}

int installed_db_manager::physical_info_callback(void *pArg, int argc, char **argv, char **columnNames) {
	physical_package_info *ph_info = (physical_package_info *)pArg;
	
	physical_file_info newfile;
	
	newfile.filename = abpmString(argv[0]);
	newfile.uid_used = 1;
	newfile.owner = mystringcpy(argv[1]);
	newfile.gid_used = 1;
	newfile.group = mystringcpy(argv[2]);
	newfile.type_used = 1;
	newfile.type = argv[3][0];
	newfile.mode_used = 1;
	newfile.mode = strtol(argv[4], NULL, 0);
	newfile.md5 = mystringcpy(argv[5]);
	newfile.major_used = 0;
	newfile.minor_used = 0;
	newfile.size_used = 1;
	newfile.size = strtol(argv[6], NULL, 0);
	newfile.time_used = 0;
	newfile.flags_used = 0;
	//FIXME: es ami meg benn van az adatbazisban?
	ph_info->insert_file(newfile);
	return 0;
}

#include "pkg_set.h"
void upgrade(Pkg_set *set, Package &pkg) {
	Package *p;
	if(set->in_name(pkg, &p)) {
		//fprintf(stderr, "%s:%d: upgrading package: %s-%s\n", __FILE__,__LINE__, pkg.name.getstring(), pkg.version->getstring());
		if(*(p->version)<*(pkg.version)) {
			set->del(*p);
			//fprintf(stderr, "%s:%d: upgradeljuk: %s-%s\n", __FILE__,__LINE__, pkg.name.getstring(), pkg.version->getstring());
			set->add(pkg);
			//set->kiir();
		} //else
			//fprintf(stderr, "megsem (%s > %s)\n", p->version->getstring(), pkg.version->getstring());
	} else {
		//fprintf(stderr, "%s:%d: adding package: %s-%s\n", __FILE__,__LINE__, pkg.name.getstring(), pkg.version->getstring());
		set->add(pkg);
	}
}


int installed_db_manager::search_required_callback(void *pArg, int argc, char **argv, char **columnNames) {
	Pkg_set *set = (Pkg_set *)pArg;
	//fprintf(stderr, "%s:%d: required package: %s-%s\n", __FILE__,__LINE__, argv[0], argv[1]);
	if(argc == 2) {
		Package pkg(argv[0], argv[1]);
		upgrade(set, pkg);
	}
	return 0;
}


bool installed_db_manager::get_physical_info(Package *pkg, physical_package_info *ph_info) {
	bool retval = true;
	char *error;
	char *pkg_name = pkg->name;
	char *pkg_version = pkg->version->getstring();
	int command_len =
		strlen(pkg_name) + 
		strlen(pkg_version) +
		200;

	char *command = new char[command_len];
	//FIXME: es ami meg benn van az adatbazisban?
	sprintf(command, "SELECT filename, file_owner, file_group, type, permission, md5, size FROM files "
			"WHERE name='%s' AND version='%s'",
			pkg_name, pkg_version);
	if(sqlite_exec(mydb, command, installed_db_manager::physical_info_callback, ph_info, &error)) {
		//fprintf(stderr, "%s\n", error);
		fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
		fprintf(stderr, "command:%s\n", command);
		retval = false;
	}
	delete[] command;
	return retval;
}

bool installed_db_manager::remove_from_db(Package *pkg) {
	bool retval = true;
	char *error;
	char *pkg_name = pkg->name;
	char *pkg_version = pkg->version->getstring();
	int command_len =
		strlen(pkg_name) + 
		strlen(pkg_version) +
		200;
	char *command = new char[command_len];
	sprintf(command, "DELETE FROM files WHERE name='%s' AND version='%s'", pkg_name, pkg_version);
	if(sqlite_exec(mydb, command, installed_db_manager::null_callback, NULL, &error)) {
		//fprintf(stderr, "%s\n", error);
		fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
		fprintf(stderr, "command:%s\n", command);
		retval = false;
	}
	sprintf(command, "DELETE FROM details WHERE name='%s' AND version='%s'", pkg_name, pkg_version);
	if(sqlite_exec(mydb, command, installed_db_manager::null_callback, NULL, &error)) {
		//fprintf(stderr, "%s\n", error);
		fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
		fprintf(stderr, "command:%s\n", command);
		retval = false;
	}
/*	sprintf(command, "DELETE FROM depends WHERE name='%s' AND version='%s'", pkg_name, pkg_version);
	sprintf(command, "DELETE FROM predepends WHERE name='%s' AND version='%s'", pkg_name, pkg_version);
	sprintf(command, "DELETE FROM provides WHERE name='%s' AND version='%s'", pkg_name, pkg_version);
	sprintf(command, "DELETE FROM conflicts WHERE name='%s' AND version='%s'", pkg_name, pkg_version);*/
	delete[] command;
	return retval;
}

static int _Find_file_callback(void *pkg_ptr, int argc, char **argv, char **columnNames) {
	Package **pkg=(Package **)pkg_ptr;

	*pkg=new Package(argv[0], argv[1]);
	//TODO: itt kene ellenorizni, h az adatbazisban nincs-e tobbszor a file
	return 0;
}

bool installed_db_manager::find_file(char *filename, Package **there) {
	Package *pkg=NULL;
	char *error;
	char *command=new char[strlen(filename)+100];
	sprintf(command, "SELECT name,version FROM files WHERE filename = '%s'", filename);
	if(sqlite_exec(mydb, command, _Find_file_callback, &pkg, &error)) {
		fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, error);
		fprintf(stderr, "command:%s\n", command);
	}
	//printf("%s:%d: command: [%s]\n",__FILE__, __LINE__, command);
	*there=pkg;
	delete[] command;
	return pkg==NULL;
}

bool database::create_installed_db(Create_mode mode, abpmString &destdir) {
	installed_db_manager dbgen;
	
	dbgen.db_init(destdir);
	dbgen.create_tables();

	if(mode != EMPTY) {
		dbgen.begin_transaction();
		dbgen.read_packages_file();
		dbgen.end_transaction();
	}
	dbgen.db_uninit();
	
	return true;
}

bool database::add_to_installed_db(package_control_info *c_info,
		physical_package_info *ph_info/*, int num*/, abpmString &destdir) {
	installed_db_manager dbgen;
	
	dbgen.db_init(destdir);
	dbgen.begin_transaction();
//	for(int i=0; i<num; i++)
	
	if(!dbgen.write_to_db(c_info->items, NULL)) {
		dbgen.db_uninit();
		return false;
	}
	if(!dbgen.write_physicalinfo_to_db(c_info, ph_info)) {
		dbgen.db_uninit();
		return false;
	}
	dbgen.commit();
	dbgen.db_uninit();
	
	return true;
}

bool database::remove_from_installed_db(Package *pkg, abpmString &destdir) {
	installed_db_manager dbgen;
	dbgen.db_init(destdir);
	dbgen.begin_transaction();
	bool retval = dbgen.remove_from_db(pkg);
	dbgen.commit();
	dbgen.db_uninit();
	return retval;
}

bool database::get_physical_info(Package *pkg, physical_package_info *ph_info, abpmString &destdir) {
	installed_db_manager dbgen;
	dbgen.db_init(destdir);
	bool retval = dbgen.get_physical_info(pkg, ph_info);
	dbgen.db_uninit();
	return retval;
}

bool database::file_conflict(char *filename, Package **p, abpmString &destdir) {
	installed_db_manager dbgen;
	dbgen.db_init(destdir);
	bool retval = dbgen.find_file(filename, p);
	dbgen.db_uninit();
	//XXX: ez igy lassu!!!!
	return retval;
}

bool database::search_required(Pkg_set *set) {
	char command[1024];
	char *error;

	sqlite *available = get_available();
	
	sprintf(command, "SELECT name, version FROM details WHERE priority='required'");
	if(sqlite_exec(available, command, installed_db_manager::search_required_callback, (void *)set, &error)) {
		fprintf(stderr, "%s\n", error);
		return false;
	}
	return true;
}

bool database::is_installed(Package *pkg, Pkg_set *inst_pkgs) {
	sqlite *installed = get_installed();
	char *error;

	abpmString command = abpmString("SELECT name, version FROM details WHERE name='") + pkg->name+
		"' AND NOT version='" + pkg->version->getstring() + "'";
	//fprintf(stderr, "%s:%d: command:%s\n", __FILE__,__LINE__,command.getstring());
	if(sqlite_exec(installed, command, db_manager::search_from_name_callback, (void *)inst_pkgs, &error)) {
		fprintf(stderr, "%s\n", error);
		return false;
	}
	if(inst_pkgs->size()>1)
		fprintf(stderr, "WARNING: more then one '%s - %s' package in the database\n",
				pkg->name.getstring(), pkg->version->getstring());
	return inst_pkgs->size() > 0;
}

