/* vi: set ts=4 sw=4: */
#ifndef __IFACE_H
#define __IFACE_H

#include "abpm_types.h"
#include <stdio.h>

#include "abpmString.h"
#include "abpmVector.h"
/*
class physical_package_info {
	public:
		int number_of_files;
		physical_file_info **files;
		
		void insert_file(physical_file_info *f_info);
//		void delete_file(int index);
		
		inline physical_package_info() {
			number_of_files=0;
			files=NULL;
		}
		~physical_package_info();
};*/

class physical_package_info:public abpmVector<physical_file_info> {
	public:
		inline void insert_file(physical_file_info &f_info) {add(f_info);}
		inline physical_package_info(): abpmVector<physical_file_info>::abpmVector() {}
		//~physical_package_info();
};

namespace physical {
	int deb_get_physical_info(char *pkg_filename, physical_package_info *info);
	int deb_get_package_info(char *pkg_filename, package_control_info *info);
	int deb_unpack(char *pkg_filename, char *prefix_dir);
	int deb_copy_scripts(char *pkg_filename, char *prefix_dir);
	int remove(physical_package_info *info, char *prefix_dir);

	void read_old_available_db(abpmString &old_db_filename, package_control_info **info, int *pkg_num);
};


class physical_iface {
	public:
		virtual bool install_package(char *path)=0;
		virtual physical_package_info get_info(char *path)=0;
		virtual bool update()=0;
};




class Pkg_set;
class Package;

namespace depend_manager {
	typedef enum {DEPEND, CONFLICT, INV_DEPEND} Depend_type;
	
	bool calculate_deps_level1(Pkg_set **install, Pkg_set **remove, abpmString destdir);
//	bool calculate_deps_level2(Pkg_set **install, Pkg_set **remove);
	bool calculate_deps_level2(Pkg_set **install, Pkg_set **remove, Pkg_set *orig_install, Pkg_set *orig_remove, abpmString destdir);
};

class media_iface {
	public:
		virtual bool get_package_file(char *name, char *version, char **path)=0;
		virtual bool get_cntl_file(char *url, char **path)=0;
};
/*
class database_iface {
	public:
		virtual bool exec_sql_command(char *command, Pkg_set *result)=0;
};
*/

//namespace physical {
	
	class Version_iface {
		public:
			virtual bool dep_ver_chk(const char *real_version)=0;
			/*this is a version shema (like ">= 1.2.3") and check if
			 * the real_version fits to this.*/
			
			virtual bool operator<(Version_iface &)=0;
			virtual bool operator>(Version_iface &)=0;
			virtual bool operator<=(Version_iface &)=0;
			virtual bool operator>=(Version_iface &)=0;
			virtual bool operator==(Version_iface &)=0;
			virtual bool operator!=(Version_iface &)=0;
			virtual void operator=(Version_iface &)=0;
			
			virtual bool operator<(const char *)=0;
			virtual bool operator>(const char *)=0;
			virtual bool operator<=(const char *)=0;
			virtual bool operator>=(const char *)=0;
			virtual bool operator==(const char *)=0;
			virtual bool operator!=(const char *)=0;
			virtual void operator=(const char *)=0;
			
			virtual Version_iface *copy()=0;
			
			virtual char *getstring()=0;
	};
	
namespace database {
	enum Create_mode {
		EMPTY,
		GENERATE
	};
	bool db_init(abpmString &destdir);
	bool db_close();
	bool create_installed_db(Create_mode mode, abpmString &destdir);
	bool update_available_db(char **packages_filenames, char **prefixes, int packages_files_count, abpmString &destdir);
	bool create_available_db(Create_mode mode, char **packages_filenames,
			char **prefixes, int packages_files_count, abpmString &destdir);
	bool add_to_installed_db(package_control_info *c_info,
			physical_package_info *ph_info/*, int num*/, abpmString &destdir);
	bool get_physical_info(Package *pkg, physical_package_info *ph_info, abpmString &destdir);
	bool remove_from_installed_db(Package *pkg, abpmString &destdir);

	bool file_conflict(char *filename, Package **p, abpmString &destdir);
	bool search_package_from_name(char *pkg_name, Pkg_set *result);
	bool search_required(Pkg_set *set);
	void search(char *search_str, abpmString &destdir);
	void show(char *pkg_name, abpmString &destdir);
	bool is_installed(Package *pkg, Pkg_set *inst_pkg);
	abpmVector<abpmString> find_urls(Package *pkg, abpmString &destdir);
};

namespace media {
	bool get_file(char *url, char **local_filename, bool with_hostname, bool force_overwrite);
	/*												^^^^^^^^^^^^^^^^^^
	 *	If this parameter (with_hostname) is true, it apperds the hostname and
	 *	directories to local filename. It is useful when using the same filenames
	 *	from different servers (eg. Packages) */
	
	
	bool get_package(abpmString& pkg_name, Version_iface *pkg_version, char **local_filename, abpmString &destdir);
//	bool get_package(Package pkg, char **local_filename);
//	TODO...

	bool update(char *config_file, abpmString destdir); /*media/parse_conffile.cpp*/
};

namespace journal {
	void begin_transaction(Pkg_set *to_install, Pkg_set *to_remove);
	void begin_package_installing(Package *pkg);
	void end_package_unpacking(Package *pkg);
	void end_package_installing(Package *pkg);
	void begin_package_removing(Package *pkg);
	void end_package_removing(Package *pkg);
	void end_transaction_normaly();
	abpmVector<abpmString> failed_transaction();
	bool check_journal();
}

#endif
