/* vi: set ts=4 sw=4: */
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include <sqlite.h>
#include <fcntl.h>
#include <pthread.h>
#include <semaphore.h>

#include "abpm_types.h"
#include "iface.h"
#include "utils.h"
//#include "install.h"
//#include "installed_db.h"
#include "version.h"
#include "pkg_set.h"
#include "package.h"
//#include "physical.h"
//#include "debug.h"
#define SERIAL_INSTALL

sem_t download_sem;
sem_t unpack_sem;
sem_t db_write_sem;

struct installing_thread_arg {
	Package *pkg;
	abpmString destdir;
};

bool run_preinst(Package *pkg, Pkg_set **inst, Pkg_set **rm, void *p_destdir);
bool run_postinst(Package *pkg, Pkg_set **inst, Pkg_set **rm, void *p_destdir);
bool run_prerm(Package *pkg, Pkg_set **inst, Pkg_set **rm, void *p_destdir);
bool run_postrm(Package *pkg, Pkg_set **inst, Pkg_set **rm, void *p_destdir);

void *_download_and_install(void *_arg) {
	physical_package_info ph_info;
	package_control_info c_info;
	struct installing_thread_arg * arg = (struct installing_thread_arg *)_arg;
	Package *pkg = arg->pkg;
	abpmString destdir = arg->destdir;
	char *pkg_filename;
#ifndef SERIAL_INSTALL	
	sem_wait(&download_sem);
#endif
	printf("downloading %s\n", pkg->name.getstring());
	media::get_package(pkg->name, pkg->version, &pkg_filename, destdir);
#ifndef SERIAL_INSTALL	
	sem_post(&download_sem);
#endif
	
	printf("installing %s\n", pkg->name.getstring());
	
	journal::begin_package_installing(pkg);
		/*TODO: ezt egyutt kene, ugy gyorsabb...*/
	
	
#ifndef SERIAL_INSTALL	
	sem_wait(&unpack_sem);
#endif
	physical::deb_copy_scripts(pkg_filename, destdir);
	run_preinst(pkg, NULL, NULL, &destdir);
	
	physical::deb_get_physical_info(pkg_filename, &ph_info);
	physical::deb_get_package_info(pkg_filename, &c_info);

	physical::deb_unpack(pkg_filename, destdir);
	journal::end_package_unpacking(pkg);
#ifndef SERIAL_INSTALL	
	sem_post(&unpack_sem);
#endif

#ifndef SERIAL_INSTALL	
	sem_wait(&db_write_sem);
#endif
	database::add_to_installed_db(&c_info, &ph_info, destdir);
#ifndef SERIAL_INSTALL	
	sem_post(&db_write_sem);
#endif
	delete pkg;
	delete arg;
	return NULL;
}

int thread_num;
pthread_t *threads;

bool do_install(Package *pkg, Pkg_set **inst, Pkg_set **rm, void *p_destdir) {
	abpmString *destdir = (abpmString *)p_destdir;

	struct installing_thread_arg * arg = new struct installing_thread_arg;
	arg->pkg = pkg->copy();
	arg->destdir = *destdir;

#ifndef SERIAL_INSTALL /*for debug*/
	if((thread_num & 1023) == 0) {
		pthread_t *new_threads = new pthread_t[thread_num + 1024];
		for(int i=0; i<thread_num; i++)
			new_threads[i] = threads[i];
		delete[] threads;
		threads = new_threads;
	}
	pthread_create(&(threads[thread_num++]), NULL, _download_and_install, (void *)arg);
#else
	_download_and_install((void *)arg);
#endif
	return true;
}

bool do_remove(Package *pkg, Pkg_set **inst, Pkg_set **rm, void *p_destdir) {
	abpmString *destdir = (abpmString *)p_destdir;

	printf("removing %s\n", pkg->name.getstring());
	run_prerm(pkg, inst, rm, p_destdir);
	physical_package_info ph_info;
	database::get_physical_info(pkg, &ph_info, *destdir);
	physical::remove(&ph_info, *destdir);
	database::remove_from_installed_db(pkg, *destdir);
	run_postrm(pkg, inst, rm, p_destdir);
	return true;
}

void usage(char *name) {
	printf("Usage:\n");
	//printf("\t%s [-D destdir] [-u] [-i package1 [package2 ...]] [-r package3 [package4 ...]]\n\n", name);
	printf("\t%s [-d destdir] [-u] [-i package1 [package2 ...]] [-r package3 [package4 ...]]\n\n", name);
	printf("Options:\n");
	printf("\t-d destdir\n\t\tuse \"destdir\" as root directory\n");
	printf("\t-u\n\t\tUpdate database before action\n");
	printf("\t-i package1 [package2 ...]]\n\t\tInstall package1 (and package2 and so on...)\n");
	printf("\t-r package1 [package2 ...]]\n\t\tRemove package1 (and package2 and so on...)\n");
}

bool create_available_db(database::Create_mode mode, abpmString destdir) {
	printf("csinalok available.db adatbazist\n");
	return media::update(destdir + SYSCONFDIR"/apt/sources.list", destdir);
}


bool check_database(abpmString db_destdir) {
	struct stat file_info;
	
	bool installed_db_need=false;
	bool available_db_need=false;
	
	abpmString filename = db_destdir + ABPM_DBDIR;
	if(!mkdir_p(filename, 0755)) {
		return false;
	}
	filename = db_destdir + ABPM_AVAILABLE_DB_FILE;
	if(stat(filename, &file_info)<0) {
		if(errno != ENOENT) {
			perror(filename);
			return false;
		} else {
			available_db_need=true;
			//fiprintf(stderr, "%s,%d:available_db_need\n", __FILE__,__LINE__);
			
		}
	}

	filename = db_destdir + ABPM_INSTALLED_DB_FILE;
	if(stat(filename, &file_info)<0) {
		if(errno != ENOENT) {
			//_debuginfo("");
			perror(filename);
			return false;
		} else {
			//TODO: itt inkabb vissza kene terni, es csak kulon kapcsoloval
			//kene adatbazist generalni...
			installed_db_need=true;
		//	fiprintf(stderr, "%s,%d:available_db_need\n", __FILE__,__LINE__);
		}
	}
	if(available_db_need) {
		if(!create_available_db(database::EMPTY, db_destdir))
			return false;
	}
	if(installed_db_need) {
		if(!create_installed_db(database::EMPTY, db_destdir))
			return false;
	}
	return true;
}
#include <stdlib.h>
#include "abpmString.h"
bool run_scripts(Package *pkg, abpmString destdir,const char *scriptname) {
	int pid = fork();
	if(pid==0) { //child process
		//fprintf(stderr, "%s:%d:running postinst script child process\n", __FILE__,__LINE__);
		chdir(destdir);
		if(chroot(".")<0) {
			//fprintf(stderr, "%s:%d:chroot nem jo\n", __FILE__,__LINE__);
			perror("chroot");
		}
		//fprintf(stderr, "%s:%d:chroot megvolt\n", __FILE__,__LINE__);
		abpmString scriptpath(ABPM_SCRIPTDIR);
		scriptpath = scriptpath + "/" + pkg->name + "_" + scriptname;
		//fprintf(stderr, "%s:%d:scriptpath = %s\n", __FILE__,__LINE__, (char *)scriptpath);
		//FILE *f=fopen(scriptpath, "rb");
		//if(f!=NULL)
		//	while(!feof(f))
		//		putchar(fgetc(f));
		//else
		//	perror("fopen");
		if(system(scriptpath)<0)
			perror("system");
		exit(0);
	}
	//fprintf(stderr, "%s:%d:running postinst script parent process\n", __FILE__,__LINE__);
	return true;
}
bool run_prerm(Package *pkg, Pkg_set **inst, Pkg_set **rm, void *p_destdir) {
	//fprintf(stderr, "%s:%d:running postinst script\n", __FILE__,__LINE__);
	abpmString *destdir=(abpmString *)p_destdir;
	return run_scripts(pkg, *destdir, "prerm.sh");
}
bool run_postrm(Package *pkg, Pkg_set **inst, Pkg_set **rm, void *p_destdir) {
	//fprintf(stderr, "%s:%d:running postinst script\n", __FILE__,__LINE__);
	abpmString *destdir=(abpmString *)p_destdir;
	return run_scripts(pkg, *destdir, "postrm.sh");
}
bool run_preinst(Package *pkg, Pkg_set **inst, Pkg_set **rm, void *p_destdir) {
	//fprintf(stderr, "%s:%d:running postinst script\n", __FILE__,__LINE__);
	abpmString *destdir=(abpmString *)p_destdir;
	return run_scripts(pkg, *destdir, "preinst.sh");
}

bool run_postinst(Package *pkg, Pkg_set **inst, Pkg_set **rm, void *p_destdir) {
	//fprintf(stderr, "%s:%d:running postinst script\n", __FILE__,__LINE__);
	abpmString *destdir=(abpmString *)p_destdir;
	bool retval = run_scripts(pkg, *destdir, "postinst.sh");
	if(retval==1)
		journal::end_package_installing(pkg);
	return retval;
}

class List {
	public:
		int num;
		char **strings;
		
		void add(char *str);
		List() {strings = NULL; num=0;}
		~List() {if(strings != NULL) delete[] strings;}
};

template <class T> T *myrealloc(T *orig_array, size_t orig_num, size_t num);

void List::add(char *str) {
	strings = myrealloc<char *>(strings, num, num + 1);
	strings[num++] = str;
}
///*
int main(int argc, char *argv[]) {
	abpmString destdir("");
	enum {INSTALL, REMOVE, SPEC_DESTDIR, NONE, SEARCH} what = NONE;
	bool need_update = false;
	//bool doing_anything = false;
	bool do_really = true;
	bool l2 = false;
	bool need_ack = true;
	bool only_search = false;
	bool only_show = false;
	char *search_str;
	List install_list, remove_list;
	
	for(int i=1; i<argc; i++) {
		if(!strcmp(argv[i], "-d")) {
			if(argc == ++i) {
				usage(argv[0]);
				return 1;
			}
			destdir = abpmString(argv[i]);
			what = NONE;
			continue;
		}
		if(!strcmp(argv[i], "-search")) {
			only_search = true;
			what = SEARCH;
			continue;
		}
		if(!strcmp(argv[i], "-show")) {
			only_show = true;
			what = SEARCH;
			continue;
		}
		if(!strcmp(argv[i], "-l2") || !strcmp(argv[i], "--level-2-check")) {
			l2 = true;
			continue;
		}
		if(!strcmp(argv[i], "-y")) {
			need_ack = false;
			continue;
		}
		if(!strcmp(argv[i], "-i")) {
			what = INSTALL;
			continue;
		}
		if(!strcmp(argv[i], "-r")) {
			what = REMOVE;
			continue;
		}
		if(!strcmp(argv[i], "-d")) {
			what = SPEC_DESTDIR;
			continue;
		}
		if(!strcmp(argv[i], "-u")) {
			need_update = true;
			continue;
		}
		if(what == INSTALL) {
			install_list.add(argv[i]);
		} else if(what == REMOVE) {
			remove_list.add(argv[i]);
		} else if(what == SPEC_DESTDIR) {
			destdir = abpmString(argv[i]);
		} else if(what == SEARCH) {
			search_str = argv[i];	
		} else {
			usage(argv[0]);
			return 1;
		}
	}
	
	if(need_update)
		if(unlink(destdir+ABPM_AVAILABLE_DB_FILE)<0) {
			if(errno != ENOENT) {
				//_debuginfo("");
				perror(destdir+ABPM_AVAILABLE_DB_FILE);
				return 2;
			}
		}
	if(!check_database(destdir))
		return 5;
	
	if(only_search) {
		database::search(search_str, destdir);
		return 0;
	}
	
	if(only_show) {
		database::show(search_str, destdir);
		return 0;
	}
	
	Pkg_set *install = new Pkg_set;
	Pkg_set *remove = new Pkg_set;
	
	//1   required csomagok felpakolasa
	database::db_init(destdir);
	database::search_required(install);
	
	//2   csomagok megkeresese
	
	for(int i=0; i<install_list.num; i++) {
		Pkg_set pkgs;
		database::search_package_from_name(install_list.strings[i], &pkgs);
		if(pkgs.size()==0) {
			fprintf(stderr, "Couldn't find package: %s\n", install_list.strings[i]);
			delete install;
			delete remove;
			database::db_close();
			return 10;
		}
		install->add(*(pkgs.getp(i)));
	}
	for(int i=0; i<remove_list.num; i++) {
		Pkg_set pkgs;
		database::search_package_from_name(remove_list.strings[i], &pkgs);
		if(pkgs.size()==0) {
			fprintf(stderr, "Couldn't find package: %s\n", remove_list.strings[i]);
			delete install;
			delete remove;
			database::db_close();
			return 11;
		}
		remove->add(*(pkgs.getp(0)));
	}
	//3   csomagok fuggosegenek ellenorzese
	
	Pkg_set *orig_install, *orig_remove;
	orig_install=install->copy(); orig_remove=remove->copy();
//	bool retval = depend_manager::calculate_deps_level1(&install, &remove);
	if(!depend_manager::calculate_deps_level1(&install, &remove, destdir)) {
		printf("Couldn't solve dependency...\n");
		do_really = false;
	}
	if(l2) {
		printf("Result of level 1 dependency checking:\n");
		printf("install:\n");
		install->kiir();
		printf("remove:\n");
		remove->kiir();
		printf("---------------------------------------\n");
		if(need_ack) {
			printf("Do you want to continue? (Y/n)\n");
			char c=fgetc(stdin);
			if(c!='y' && c!='Y' && c!='\n') {
				printf("aborting...\n");
				do_really=false;
			}
			while(c!='\n')
				c=fgetc(stdin);
		}

		if(!depend_manager::calculate_deps_level2(&install, &remove, orig_install, orig_remove, destdir)) {
			printf("Couldn't solve dependency...\n");
			do_really = false;
		}
	}
	//3.1 a leszedesre kerulo csomagokat megprobaljuk upgrade-del "helyrepofozni"
	// (a nagyobb verzioju csomagokat felvesszuk az install halmazba)
	//TODO...
	
	printf("install:\n");
	install->kiir();
	printf("remove:\n");
	remove->kiir();
	if(do_really) {
		printf("---------------------------------------\n");
		if(need_ack) {
			printf("Are you sure? (Y/n)\n");
			char c=fgetc(stdin);
			if(c!='y' && c!='Y' && c!='\n') {
				printf("aborting...\n");
				do_really=false;
			}
			while(c!='\n')
				c=fgetc(stdin);
		}
	}
	if(!do_really) {
		delete install;
		delete remove;
		database::db_close();
		return 0;
	}
	//4   csomagok letoltese, felpakolasa/leszedese
#ifndef SERIAL_INSTALL
	sem_init(&download_sem, 0, 1);
	sem_init(&unpack_sem, 0, 1);
	sem_init(&db_write_sem, 0, 1);
	if(retval) {
#endif
		thread_num = 0;
#ifndef SERIAL_INSTALL
		sem_wait(&unpack_sem);
		sem_wait(&db_write_sem);
#endif
		remove->foreach(do_remove, &install, &remove, (void *)&destdir);
		install->foreach(do_install, &install, &remove, (void *)&destdir);
#ifndef SERIAL_INSTALL
		sem_post(&db_write_sem);
		sem_post(&unpack_sem);
	}
	
	for(int i=0; i<thread_num; i++)
		pthread_join(threads[i], NULL);
	
	sem_destroy(&download_sem);
	sem_destroy(&unpack_sem);
	sem_destroy(&db_write_sem);	
#endif
	install->foreach(run_postinst, &install, &remove, (void *)&destdir);
	delete install;
	delete remove;
	database::db_close();
	return 0;
}
/* /

#include "abpmVector.h"
#include "pkg_set.h"

bool f(Package *i, void *e) {
	printf("%s\n", i->name.getstring());
	return true;
}

int main(int argc, char *argv[]) {

	Pkg_set s;
	Package *p = new Package("aoeu", "1.2");
	s.add(p);
	s._foreach(f,NULL);
	return 0;
}
*/
