/* vi: set ts=4 sw=4: */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>

#include "downloader.h"
#include "utils.h"

#define _debuginfo(info) fprintf(stderr, "%s: %d: %s", __FILE__, __LINE__, info)

bool downloader::tcp_connect(int port) {
	struct sockaddr_in addr;
	struct hostent* hostinfo;
	
	if(!is_file("/etc/resolv.conf")) {
		perror("/etc/resolv.conf");
		return false;
	}
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if(sock < 0) {
		//fprintf(stderr,"%s: tcp_connec(): 28: ");
		_debuginfo("");
		perror("socket");
		close(out_fd);
		return false;
	}
	
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	hostinfo = gethostbyname(host);

	if(hostinfo==NULL) {
		//fprintf(stderr,"downloader.cpp: tcp_connec(): 39: ");
		_debuginfo("gethostbyname returned NULL");
		perror(host);
		close(out_fd);
		return false;
	}
	
	addr.sin_addr = *(struct in_addr *)hostinfo->h_addr;
	
	if(connect(sock, (struct sockaddr *)&addr, sizeof(addr))<0) {
		//fprintf(stderr,"downloader.cpp: tcp_connec(): 48: ");
		_debuginfo("");
		perror("connect");
		close(out_fd);
		return false;
	}
	return true;
}

bool downloader::tcp_connect_ip4(char *ip, int port) {
	struct sockaddr_in addr;
	in_addr_t ip_addr;
	
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if(sock < 0) {
		//fprintf(stderr,"downloader.cpp: tcp_connec_ip4(): 62: ");
		_debuginfo("");
		perror("socket");
		close(out_fd);
		return false;
	}
	
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	
	ip_addr = inet_addr(ip);
	
	addr.sin_addr = *((struct in_addr *)&ip_addr);
	
	if(connect(sock, (struct sockaddr *)&addr, sizeof(addr))<0) {
		//fprintf(stderr,"downloader.cpp: tcp_connec_ip4(): 76: ");
		_debuginfo("");
		perror("connect");
		close(out_fd);
		return false;
	}
	//printf("downloader.cpp: tcp_connect_ip4(): 75: sock = %d cmd_sock = %d\n", sock, cmd_sock);
	return true;
}

bool downloader::tcp_connect_ip6(char *ip, int port) {
	fprintf(stderr, "Warning: Sorry, ipv6 is not implemented yet. :-(\n");
	return false;
}

bool downloader::sendall(int socket, char *buf, size_t bytes) {
	size_t sent_bytes = 0;
	size_t act_sent_bytes;
	//fprintf(stderr, "downloader::sendall(): elkuldve:%s\n", buf);
	//printf("downloader::sendall(): elkuldve:%s\n", buf);
	while(sent_bytes < bytes) {
		act_sent_bytes = send(socket, buf+sent_bytes, bytes - sent_bytes, 0);
		if(act_sent_bytes <= 0) {
			//fprintf(stderr,"downloader.cpp: sendall(): 98: ");
			_debuginfo("");
			perror("send");
			return false;
		}
		sent_bytes += act_sent_bytes;
	}
	return true;
}

bool downloader::recv_data_to_file() {
	char buf[1024];
	ssize_t recv_bytes=0, act_recv_bytes;
	ssize_t written_bytes, act_written_bytes;
	
	//printf("recv_data_to_file(%d)\nout_fd = %d\n", length, out_fd);
	
	fd_set read_fds;
	fd_set infds;
	
	int max_sock = sock > cmd_sock ? sock+1 : cmd_sock+1;
	
	FD_ZERO(&read_fds);
	FD_SET(sock, &read_fds);
	if(cmd_sock > 0 )
		FD_SET(cmd_sock, &read_fds);
	FD_ZERO(&infds);
	while(length < 0 || recv_bytes < length) {
		//printf("downloader.cpp: recv_data_to_file(): 118: select jon max_sock = %d\n", max_sock);
		infds = read_fds;
		select(max_sock, &infds, NULL, NULL, NULL);

		//printf("downloader.cpp: recv_data_to_file(): 121: select visszatert\n");

		if(!FD_ISSET(sock, &infds)) {
			//meg egy proba
			sleep(1);
			infds = read_fds;
			select(max_sock, &infds, NULL, NULL, NULL);

			if(!FD_ISSET(sock, &infds)) {
				//printf("downloader.cpp: recv_data_to_file(): 124: a cmd_sock kapta meg a vezerlest\n");
				/*if(!FD_ISSET(cmd_sock, &infds)) {
				  printf("downloader.cpp: recv_data_to_file(): 127: cmd_sock siscs benne...\n");
				  } else {
				  int l = recv(cmd_sock, buf, 1024, 0);
				  buf[l] = '\0';
				  printf(buf);
				  }*/
				close(sock);
				close(out_fd);
				if(length < 0)
					return true;
				else
					return false;
			}
		}

		act_recv_bytes = recv(sock, buf, 1024, 0);
		if(act_recv_bytes <0) {
			//fprintf(stderr,"%s: recv_data_to_file(): %d: ", __FILE__,__LINE__);
			_debuginfo("");
			perror("recv");
			close(sock);
			close(out_fd);
			return false;
		}
		if(act_recv_bytes == 0) {
			if(cmd_sock <0) {
				if(length < 0)
					return true;
				else {
					close(sock);
					close(out_fd);
					return false;
				}
			} else {
				if(!FD_ISSET(cmd_sock, &infds)) {
					infds = read_fds;
					sleep(1);
					select(max_sock, &infds, NULL, NULL, NULL);
					/*if(!FD_ISSET(cmd_sock, &infds)) {
						fprintf(stderr,"%s: %d: errno = %d\n", __FILE__, __LINE__, errno);
						_debuginfo("");
						//fprintf(stderr,"%s: recv_data_to_file(): %d: ",__FILE__,__LINE__);
						perror("recv");
						//continue;
					}*/
					if(close(sock) < 0) {
						_debuginfo("");
						perror("close");
					}
					if(close(out_fd)<0) {
						_debuginfo("");
						perror("close");
					}
					if(FD_ISSET(cmd_sock, &infds))
						return true;
					else
						return false;
				}
			}
		}
		written_bytes = 0;
		while(written_bytes < act_recv_bytes) {
			act_written_bytes = write(out_fd, buf + written_bytes, act_recv_bytes - written_bytes);
			if(act_written_bytes <= 0) {
				//printf("itt szall el. out_fd = %d\n", out_fd);
				//fprintf(stderr,"%s: recv_data_to_file(): %d: ",__FILE__, __LINE__);
				_debuginfo("");
				perror("write");
				close(out_fd);
				close(sock);
				return false;
			}
			written_bytes += act_written_bytes;
		}
	}
	close(sock);
	close(out_fd);
	return true;
}

bool downloader::create_tmp_file(bool can_append) {
	
	char *dir = dirof(real_filename);
	bool retval = mkdir_p(dir, 0755);
	if(!retval) {
	//	printf("itt szallunk ki. dir = %s\n", dir);
		delete[] dir;
		return false;
	}
	delete[] dir;
	
	int len = strlen(real_filename);
	tmp_filename = new char[len + 10];
	strcpy(tmp_filename, real_filename);
	strcpy(tmp_filename+len, ".part");
	
	if(can_append) {
		out_fd = open(tmp_filename, O_WRONLY|O_CREAT|O_APPEND, 0644);
		offset = file_size(real_filename);
	} else {
		out_fd = open(tmp_filename, O_WRONLY|O_CREAT, 0644);
		offset = 0;
	}
	
	if(out_fd <= 0) {
		//fprintf(stderr,"downloader.cpp: create_tmp_file(): 215: ");
		_debuginfo("");
		perror(tmp_filename);
		delete[] tmp_filename;
		tmp_filename = NULL;
		return false;
	}

	return true;
}

bool downloader::move_to_real_file() {
	//printf("moving %s to %s\n", tmp_filename, real_filename);
	if(unlink(real_filename)<0)
		if(errno != ENOENT) {
			//fprintf(stderr,"downloader.cpp: move_to_real_file(): 229: ");
			_debuginfo("");
			perror(real_filename);
			return false;
		}
	if(link(tmp_filename, real_filename)<0) {
		//fprintf(stderr,"downloader.cpp: move_to_real_file(): 234: ");
		_debuginfo("");
		perror(real_filename);
		return false;
	}
	if(unlink(tmp_filename)<0) {
		//fprintf(stderr,"downloader.cpp: move_to_real_file(): 239: ");
		_debuginfo("");
		perror(tmp_filename);
		return false;
	}
	return true;
}
