/* vi: set ts=4 sw=4: */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
//#include <pthread.h>

#include "ftp_downloader.h"
#include "iface.h"
#include "utils.h"

using namespace media;

#define FTP_PORT 21


char net_reader::get_char() {
	int act_read_bytes;

	if(_eof)
		return EOF;

	if(recv_bytes <= read_bytes) {
		act_read_bytes = recv(fd, buf, 1024, 0);
		if(act_read_bytes <= 0) {
			if(act_read_bytes != 0) {
				//fprintf(stderr,"ftp_downloader.cpp: get_char(): 31: ");
				//_debuginfo("");
				perror("recv");
			}
			_eof = true;
			return EOF;
		}
		read_bytes = 0;
		recv_bytes = act_read_bytes;
	}
	return buf[read_bytes++];
}

template<class T> T *myrealloc(T *orig_array, size_t orig_num, size_t num);

char *net_reader::next_line() {
	int result_mem = 1024;
	int len = 0;

	char *result = new char[result_mem];
	do {
		if(len >= result_mem - 1) {
			result = myrealloc<char>(result, result_mem, result_mem+1024);
			result_mem += 1024;
		}
		result[len++] = get_char();
	} while(result[len-1] != '\n' && !_eof);
	result[len - 1] = '\0';
	//	printf("len = %d\n", len);
	return result;
}

int ftp_downloader::recv_result_line() {
	//int read_bytes = 0;
	//int act_read_bytes;
	//int result_len = strlen(expected_result);
	//	int result_len = 3;
	//int newline=0;

	bool lastline=false;
	//bool halfline=false;
	bool firstline=true;
	int cmd_num = 0;
	char cmd_id[4];


	//printf("ftp_result:\n");
	while(!lastline) {
		char *line = nr->next_line();
		if(firstline) {
			strncpy(cmd_id, line, 3);
			cmd_id[3] = '\0';
			sscanf(cmd_id, "%d", &cmd_num);
			firstline = false;
		}
		if(strncmp(line, cmd_id, 3) == 0 && line[3] != '-') {
			lastline = true;
		}
		delete[] line;
	}
	return cmd_num;
}

bool ftp_downloader::protocol(char *filename) {
	int cmd_num;

	if(recv_result_line() != 220)
		return false;

	if(!sendall(cmd_sock, "USER anonymous\n", 15))
		return false;

	cmd_num = recv_result_line();
	if(cmd_num == 331) {
		if(!sendall(cmd_sock, "PASS -abpm@\n", 12))
			return false;

		cmd_num = recv_result_line();
	}
	if(cmd_num != 230)
		return false;


	//if(!sendall(cmd_sock, "PWD\n", 4))
	//	return false;
	//Abszolut path-ot adunk meg...

	if(!sendall(cmd_sock, "TYPE I\n", 7))
		return false;

	if(recv_result_line() != 200)
		return false;

	char *dir = dirof(filename);
	sprintf(buf, "CWD /%s\n", dir);
	delete[] dir;

	if(!sendall(cmd_sock, buf, strlen(buf)))
		return false;

	if(recv_result_line() != 250)
		return false;

	if(!sendall(cmd_sock, "PASV\n", 5))
		return false;

	if(!create_data_sock())
		return false;

	//printf("ftp_downloader.cpp: 123: filename: [%s]\n", filename);
	char *bname = basename(filename);
	//printf("ftp_downloader.cpp: 125: basename(filename): [%s]\n", bname);
	sprintf(buf, "RETR %s\n", bname);

	if(!sendall(cmd_sock, buf, strlen(buf)))
		return false;

	//TODO: a fajl hosszat meg kene tudni mondani...
	if(recv_result_line() != 150)
		return false;

	return true;
}

bool ftp_downloader::get_important_string(char *expected_start, char **imp_str) {
	ssize_t read_bytes=0, act_read_bytes;

	int result_start, result_end;
	/*FIXME: mi van, ha tobb sort kuldenek???*/
	while(read_bytes < 4) {
		//Minden bizonnyal ez a ciklus csak 1x fut le :-)

		act_read_bytes = recv(cmd_sock, buf + read_bytes, BUF_SIZE - read_bytes, 0);
		if(act_read_bytes <= 0) {
			//fprintf(stderr,"ftp_downloader.cpp: get_important_string(): 164: ");
			//_debuginfo("");
			perror("recv");
			return false;
		}
		read_bytes += act_read_bytes;
	}
	if(strncmp(buf, expected_start, 3)) {
		buf[read_bytes] = '\0';
		fprintf(stderr, "ftp connection error:\n%s\n", buf);
		return false;
	}
	while(1) {
		int i;
		for(i=0; i<read_bytes && buf[i] != '('; i++);
		if(buf[i] !='(') {
			read_bytes = recv(cmd_sock, buf, BUF_SIZE, 0);
			if(read_bytes <= 0) {
				//fprintf(stderr,"ftp_downloader.cpp: get_important_string(): 181: ");
				//_debuginfo("");
				perror("recv");
				return false;
			}
			continue;
		}
		int j;
		for(j=i; j<read_bytes && buf[j] != ')'; j++);
		if(buf[j] !=')') {
			for(int k=i; k<read_bytes; k++)
				buf[k-i] = buf[k];
			i=read_bytes - i;
			read_bytes = recv(cmd_sock, buf + i, BUF_SIZE - i, 0);
			if(read_bytes <= 0) {
				//fprintf(stderr,"ftp_downloader.cpp: get_important_string(): 195: ");
				//_debuginfo("");
				perror("recv");
				return false;
			}
			read_bytes += i;
			continue;
		}
		result_start = i;
		result_end = j;
		break;
	}
	char *result = new char[result_end - result_start + 1];
	result_start++;
	int i;
	for(i=result_start; i<result_end; i++)
		result[i - result_start] = buf[i];
	result[i - result_start] = '\0';
	*imp_str = result;
	return true;
}

bool ftp_downloader::start_download() {
	bool result;
	char *len_str;
	result = get_important_string("150", &len_str);
	if(result)
		length = strtol(len_str, NULL, 0);
	/*FIXME: valami ellenorzes kene...*/

	return result;
}


bool ftp_downloader::create_data_sock() {
	bool result;
	char *ip;
	result = get_important_string("227", &ip);
	int comas = 0;
	int i;
	//printf("ftp_downloader.cpp: create_data_sock(): 231: rendes socketet csinalunk. IP=%s\n", ip);
	for(i=0; comas < 4 && ip[i] != '\0'; i++) {
		if(ip[i] == ',') {
			comas++;
			if(comas == 4)
				ip[i] = '\0';
			else
				ip[i] = '.';
		}
	}

	char *nextnum;
	int port = strtol(ip + i, &nextnum, 0) * 256;
	port += strtol(nextnum + 1, NULL, 0);
	/*FIXME: valami ellenorzes kene...*/

	//printf("ftp_downloader.cpp: create_data_sock(): 231: rendes socketet csinalunk. IP=%s, port=%d\n", ip, port);
	if(result) {
		result = tcp_connect_ip4(ip, port);
		/*FIXME: mi lesz az ipv6-tal????*/
		delete[] ip;
	}
	return result;
}

bool ftp_downloader::finish_connection() {
	bool result = true;

	//printf("ftp_downloader.cpp: finish_connection(): 259: indul\n");
	if(recv_result_line() != 226)
		result = false;

	//printf("ftp_downloader.cpp: finish_connection(): 263: recv_result_line() megvolt\n");
	/*if(close(out_fd) < 0) {
		//fprintf(stderr,"ftp_downloader.cpp: finish_connection(): 268: ");
		_debuginfo("");
		perror("close");
		result = false;
	}*/

	//printf("ftp_downloader.cpp: finish_connection(): 269: close(out_fd) megvolt\n");
/*	if(close(sock) < 0) {
		//fprintf(stderr,"ftp_downloader.cpp: finish_connection(): 275: ");
		_debuginfo("");
		perror("close");
		result = false;
	}*/
	//printf("ftp_downloader.cpp: finish_connection(): 274: close(sock) megvolt\n");

	if(close(cmd_sock) < 0) {
		//fprintf(stderr,"ftp_downloader.cpp: finish_connection(): 282: ");
		//_debuginfo("");
		perror("close");
		result = false;
	}
	//printf("ftp_downloader.cpp: finish_connection(): 280: close(cmd_sock) megvolt\n");

	return result;
}
/*
   void *ftp_downloader::_data_download_thread(void *classptr) {
   ftp_downloader *owner = (ftp_downloader *)classptr;

   owner->_thread_downloader_fun();
   return NULL;
   }*/
/*
   void ftp_downloader::_thread_downloader_fun() {
   download_result = recv_data_to_file();
   }
   */
	bool ftp_downloader::_download(char *file) {
		if(!create_tmp_file(false))
			return false; //^^^^^ TODO: ezt lehetne finoman hangolni...

		bool retval = tcp_connect(FTP_PORT);
		if(!retval)
			return false;
		cmd_sock = sock;


		length = -1;

		nr = new net_reader(cmd_sock);

		if(!protocol(file)) {
			delete nr;
			return false;
		}

		//printf("ftp_downloader.cpp: _download(): 310: indul a tenyleges letoltes\n");
		retval = recv_data_to_file();
		//	if(retval)
		retval = finish_connection();
		//if(!retval)
			//_debuginfo("gond van...\n");
		retval = move_to_real_file();
		//_debuginfo("atneveztuk\n");
		return retval;
	}
/*
   bool ftp_downloader::download(char *hostname, char *file, char *out_filename) {
   ftp_downloader dl(hostname);
   return dl._download(file, out_filename);
   }*/

bool ftp_downloader::download(char *hostname, char *file, char *out_filename) {
	downloader_thread_arg *arg = new downloader_thread_arg;

	arg->host = new char[strlen(hostname)+1];
	strcpy(arg->host, hostname);
	arg->file = new char[strlen(file)+1];
	strcpy(arg->file, file);
	arg->out_filename = new char[strlen(out_filename)+1];
	strcpy(arg->out_filename, out_filename);

	//pthread_create...
	downloader_thread((void *)arg);
	//_debuginfo("lejott\n");
	return true;
}

void *ftp_downloader::downloader_thread(void *_arg) {
	downloader_thread_arg *arg = (downloader_thread_arg *)_arg;

	ftp_downloader dl(arg->host, arg->out_filename);

	//printf("419 hostname:%s\n", arg->host);
	dl._download(arg->file);
	delete[] arg->host;
	delete[] arg->file;
	delete[] arg->out_filename;
	delete arg;
	//_debuginfo("downloader_thread vege\n");
	return NULL;
}

