/* vi: set ts=4 sw=4: */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include "downloader.h"
#include "http_downloader.h"
#include "ftp_downloader.h"
#include "get_file.h"
#include "utils.h"

//#define ABPM_CACHEDIR "."

char * media::tmp_filename(char *orig_filename) {
	char *bname = basename(orig_filename);
	char *result = new char[strlen(bname) + strlen(ABPM_CACHEDIR) + 1];
	sprintf(result, "%s/%s", ABPM_CACHEDIR, bname);
	delete[] bname;
	//return destdir + ABPM_CACHEDIR + "/" + basename(orig_filename);
}

bool media::get_file(char *url, char **local_filename, bool with_hostname, bool force_overwrite) {
	char *host;
	char *filename;
	Protocol_type proto;
	bool retval=false;
	int magic_len;
	
	int url_len = strlen(url);
	
	
	if(!strncmp(url, "http://", 7)) {
		magic_len = 7;
		proto=HTTP;
	} else if(!strncmp(url, "ftp://", 6)) {
		magic_len = 6;
		proto=FTP;
	} else if(!strncmp(url, "rsync://", 8)) {
		magic_len = 8;
		proto=RSYNC;
	} else if(!strncmp(url, "cdrom:", 6)) {
		magic_len = 6;
		proto=CDROM;
	} else if(!strncmp(url, "file:", 5)) {
		magic_len = 5;
		proto=FILE_SYSTEM;
	} else {
		fprintf(stderr, "unknown type of URL: %s\n", url);
		*local_filename = NULL;
		return false;
	}
	
	//printf("%s: get_file(): %d: url:%s\n", __FILE__, __LINE__, url);
	
	host = new char[url_len];
	strncpy(host, url + magic_len, url_len - magic_len + 1);
														   /*^^^*/
									/*az utolso 0-t is be kell masolni*/
	for(filename = host; filename[0]!='/'; filename++);
	filename[0] = '\0';
	filename++;
	int cachedir_len = strlen(ABPM_CACHEDIR);
	char *out_filename = new char[cachedir_len + url_len + 2];
	strcpy(out_filename, ABPM_CACHEDIR);
	out_filename[cachedir_len] = '/';
	if(with_hostname) {
		int i;
		for(i = 0; i < url_len; i++)
			out_filename[i + cachedir_len + 1] = (url[i] == '/' || url[i] == ':') ? '_' : url[i];
		out_filename[i + cachedir_len + 1] = '\0';
	}
	else
		strcpy(out_filename + cachedir_len + 1, filename);
	//printf("79: filename = [%s]\n", filename);
	
	if(force_overwrite)
		if(unlink(out_filename)<0)
			if(errno != ENOENT) {
				//_debuginfo("");
				perror(out_filename);
			}

	if(!is_file(out_filename)) {
	//printf("%s: %d: out_filename = [%s]\n", __FILE__, __LINE__, out_filename);
		switch(proto) {
			case HTTP:
				retval = http_downloader::download(host, filename, out_filename);
				*local_filename = out_filename;
				break;
			case FTP:
				//printf("91: out_filename = [%s]\n", out_filename);
				retval = ftp_downloader::download(host, filename, out_filename);
				*local_filename = out_filename;
				break;
			default:
				fprintf(stderr, "Warning: protocol of URL (%s) is not implemented yet\n", url);
				delete[] out_filename;
				delete[] host;
//				*local_filename = NULL;
				return false;
		}
	} else {
		//printf("get_file.cpp: 108: a file mar letezik\n");
		*local_filename = out_filename;
		retval = true;
	}
	delete[] host;
	return retval;
}


