/* vi: set ts=4 sw=4: */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "http_downloader.h"
#include "iface.h"
#include "utils.h"

using namespace media;


#define HTTP_PORT 80

bool http_downloader::send_command(char *file, size_t offset) {
	char *command = new char[strlen(file) + 200];
	
	if(offset == 0)
		sprintf(command, "GET /%s HTTP/1.0\n"
				"User-Agent: abpm\n"
				"Host: %s\n"
				"Connection: Keep-Alive\n\n",
				file, host);
	else
		sprintf(command, "GET /%s HTTP/1.0\n"
				"User-Agent: abpm\n"
				"Accept: */*\n"
				"Host: %s\n"
				"Connection: Keep-Alive\n"
				"Range: bytes=%d-\n\n",
				file, host, offset);
	//printf("%s %d host=%s\n",__FILE__, __LINE__, host);
	
	int command_len = strlen(command);
	int sent_bytes = 0;
	int act_sent_bytes;
	while(sent_bytes < command_len) {
		act_sent_bytes = send(sock, command + sent_bytes, command_len - sent_bytes, 0);
		if(act_sent_bytes <= 0) {
			perror("send");
			delete[] command;
			return false;
		}
		sent_bytes += act_sent_bytes;
	}
	delete[] command;
	return true;
}

bool http_downloader::recv_header(int *processed_bytes) {
	length = -1;
	int line_offset = 0;
	while(1) {
		int line_start = 0;
		int line_end = 0;
		bytes_in_buf = recv(sock, buf + line_offset, BUF_SIZE, 0);
		while(1) {
			for(line_end = line_start; line_end < bytes_in_buf && buf[line_end] != '\n'; line_end++);
			if(line_end == bytes_in_buf &&
					(buf[line_end] != '\n' || line_end < 0 || buf[line_end - 1] != '\n'))
				break;
			if(!strncmp("\r\n", buf+line_start, 2) || line_end == line_start) {
				*processed_bytes = line_end + 1;
				return true;
			}
			if(!strncmp(buf+line_start, "Content-Length: ", 16))
				length = strtol(buf+line_start+16, NULL, 0);
			line_start = line_end+1;
			/*FIXME Es a hibakezeles??? (pl error404)*/
		}
		/*az utolso felsort a buffer elejere masoljuk. Ha az elejen van, akkor
		 * hagyjuk a francba, mert 1024-nel hosszabb sor erkezett.*/
		if(line_start==0) {
			fprintf(stderr, "WARNING: wrong http header\n");
			line_offset = 0;
			continue;
		}
		for(int i=line_start; i<line_end; i++)
			buf[i-line_start] = buf[i];
		line_offset = line_end - line_start;
	}
	return false; /*ide nem juthatunk el...*/
}

bool http_downloader::_download(char *file, char *out_file) {
	int out_fd;
	ssize_t recv_bytes = 0;
	ssize_t act_recv_bytes;
//	ssize_t written_bytes;
	ssize_t act_written_bytes;
	
	out_fd = open(out_file, O_WRONLY|O_CREAT|O_APPEND, 0644);
	if(out_fd <0) {
		perror(out_file);
		return false;
	}
	
//	printf("%s: %d: downloading: %s\n", __FILE__, __LINE__, file);
	if(!tcp_connect(HTTP_PORT))
		return false;
	printf("%s: %d: downloading: %s\n", __FILE__, __LINE__, file);
	size_t offset = file_size(out_file);
	if(!send_command(file, offset)) {
		close(out_fd);
		close(sock);
		return false;
	}
//	printf("%s: %d: downloading: %s\n", __FILE__, __LINE__, file);
	int processed_bytes;
	if(!recv_header(&processed_bytes)) {
		close(out_fd);
		close(sock);
		return false;
	}
//	printf("%s: %d: downloading: %s\n", __FILE__, __LINE__, file);
	act_recv_bytes = bytes_in_buf - processed_bytes;
	if(length<0) {
		close(out_fd);
		close(sock);
		return false;
	}
//	printf("%s: %d: downloading: %s\n", __FILE__, __LINE__, file);
	while(1) {
		//TODO na ezt szebben kene...
		recv_bytes += act_recv_bytes;
		while(act_recv_bytes > 0) {
			act_written_bytes = write(out_fd, buf+processed_bytes, act_recv_bytes);
			if(act_written_bytes <= 0) {
				perror("write");
				close(out_fd);
				close(sock);
				return false;
			}
			processed_bytes += act_written_bytes;
			act_recv_bytes -= act_written_bytes;
		}
		processed_bytes = 0;
		if(recv_bytes >= length)
			break;
		act_recv_bytes = recv(sock, buf, BUF_SIZE, 0);
		if(act_recv_bytes <= 0) {
			if(act_recv_bytes <0)
				perror("recv");
			close(out_fd);
			close(sock);
			if(recv_bytes <0)
				return false;
			return false;
		}
	}
//	printf("%s: %d: downloading: %s\n", __FILE__, __LINE__, file);
	return true;
}

void _urlcpy(char *dst, char *src) {
	int i, j=0;
	for(i=0; src[i]!=0; i++) {
		if(src[i]!='/') {
			dst[j] = src[i];
			j++;
			continue;
		}
		while(src[i+1]=='/')
			i++;
		if(src[i+1]=='.' && src[i+2]=='/')
			i++;
		while(src[i+1]=='/')
			i++;
		dst[j]='/';
		j++;
	}
	dst[j]=0;
}

bool http_downloader::download(char *hostname, char *file, char *out_filename) {
	downloader_thread_arg *arg = new downloader_thread_arg;
	
	printf("%s: %d: downloading: %s\n", __FILE__, __LINE__, file);
	
	arg->host = new char[strlen(hostname)+1];
	strcpy(arg->host, hostname);
	arg->file = new char[strlen(file)+1];
	_urlcpy(arg->file, file);
	arg->out_filename = new char[strlen(out_filename)+1];
	strcpy(arg->out_filename, out_filename);

	//pthread_create...
	downloader_thread((void *)arg);
	return true;
}

void *http_downloader::downloader_thread(void *_arg) {
	downloader_thread_arg *arg = (downloader_thread_arg *)_arg;
		
	http_downloader dl(arg->host);
	dl._download(arg->file, arg->out_filename);
	delete[] arg->host;
	delete[] arg->file;
	delete[] arg->out_filename;
	delete arg;
	return NULL;
}
