/* vi: set ts=4 sw=4:*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
//#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <utime.h>
#include "abpmArFile.h"
#include "abpmRawFile.h"
#include "abpmWrongFile.h"
#include "abpmVector.h"
#include "abpmString.h"

#define ARMAG "!<arch>\n"	/* magic string */
#define SARMAG 8		/* length of magic string */
#define ARFMAG "`\n"		/* header trailer string */

struct ar_hdr {			/* file member header */
	char ar_name[16];	/* '/' terminated file member name */
	char ar_date[12];	/* file member date */ 
	char ar_uid[6];		/* file member user identification */ 
	char ar_gid[6];		/* file member group identification */ 
	char ar_mode[8];	/* file member mode (octal) */ 
	char ar_size[10];	/* file member size */ 
	char ar_fmag[2];	/* header trailer string */ 
}; 


abpmArFile::abpmArFile(abpmString filename, bool useraw) {
	file_choosen = false;
	use_raw=useraw;
	raw_reader = new abpmVirtualFile(filename);
}

abpmArFile::abpmArFile(abpmFile *rawreader, bool useraw) {
	file_choosen = false;
	use_raw=useraw;
	raw_reader = rawreader;
}

abpmArFile::abpmArFile(abpmFile *rawreader, abpmString subfile_path, bool useraw) {
	raw_reader = rawreader;
	use_raw=useraw;
	choose_file(subfile_path);
}

off_t abpmArFile::seek(off_t offset, int whence) {
	if(!file_choosen)
		return 0;
	off_t absolute_offset = start + offset; //SEEK_SET
	if(whence==SEEK_CUR)
		absolute_offset = raw_reader->seek(0, SEEK_CUR) + offset;
	if(offset==SEEK_END)
		absolute_offset = end + offset;
	if(absolute_offset > end)
		absolute_offset = end;
	if(absolute_offset < start)
		absolute_offset = start;
	off_t o_offset = raw_reader->seek(absolute_offset, SEEK_SET);
	return o_offset - start;
}

bool abpmArFile::eof() {
	if(!file_choosen)
		return true;
	off_t cur_pos = raw_reader->seek(0, SEEK_CUR);
	if(cur_pos >= end || cur_pos < start)
		return true;
	return false;
}

abpmVector<abpmString> abpmArFile::list() {
	abpmVector<abpmString> contents;
	struct ar_hdr header;
	char size_str[11];
	size_t size;
	int i;
	char c;
	
	off_t o_offset = raw_reader->seek(0, SEEK_CUR);
	raw_reader->seek(0,SEEK_SET);
	if(!check_magic()) {
		fprintf(stderr, "Warning!! It's not 'ar' file format!\n");
		raw_reader->seek(o_offset,SEEK_SET);
		return contents;
	}

//	filename[16] = '\0';
	size_str[10] = '\0';
	while(readheader((char *)(&header))) {
		memcpy(size_str, header.ar_size, 10);
		//memcpy(filename, header.ar_name, 16);
		for(i=15; i>0; i--) {
			if(header.ar_name[i] == ' ')
				header.ar_name[i] = '\0';
			else
				break;
		}
		size = strtol(size_str, NULL, 0);
		/**/
		abpmString ar_name(header.ar_name);
		contents.add(ar_name);
		raw_reader->seek(size, SEEK_CUR);
		if(raw_reader->read(&c, 1) > 0 && c != '\n')
			raw_reader->seek(-1, SEEK_CUR);
	}
	raw_reader->seek(o_offset,SEEK_SET);
	return contents;
}

bool abpmArFile::is_file(abpmString path) {
	struct ar_hdr header;
	char size_str[11];
	size_t size;
	int i;
	char c;
	
	for(i=0; path[i]=='/'; i++);
	abpmString basename=path.tail(i);
	
	off_t o_offset = raw_reader->seek(0, SEEK_CUR);
	raw_reader->seek(0,SEEK_SET);
	if(!check_magic()) {
		fprintf(stderr, "Warning!! It's not 'ar' file format!\n");
		raw_reader->seek(o_offset,SEEK_SET);
		return false;
	}

//	filename[16] = '\0';
	size_str[10] = '\0';
	while(1) {
		if(!readheader((char *)(&header)))
			break;
		memcpy(size_str, header.ar_size, 10);
		//memcpy(filename, header.ar_name, 16);
		for(i=15; i>0; i--) {
			if(header.ar_name[i] == ' ')
				header.ar_name[i] = '\0';
			else
				break;
		}
		size = strtol(size_str, NULL, 0);
		/**/
		if(!strncmp(header.ar_name, basename, 16)) {
			start = raw_reader->seek(0, SEEK_CUR);
			end = start + size;
			raw_reader->seek(o_offset,SEEK_SET);
			return true;
		}
		raw_reader->seek(size, SEEK_CUR);
		if(raw_reader->read(&c, 1) > 0 && c != '\n')
			raw_reader->seek(-1, SEEK_CUR);
	}
	raw_reader->seek(o_offset,SEEK_SET);
	return false;
}

ssize_t abpmArFile::read(void *buf, size_t maxlen) {
	size_t real_maxlen = maxlen;
	off_t cur_pos = raw_reader->seek(0, SEEK_CUR);
	if(cur_pos >= end)
		return 0;
	if((off_t)(cur_pos + maxlen) > end)
		real_maxlen = end - cur_pos;
	return raw_reader->read(buf, real_maxlen);
}



/*returns a file descriptor that seeked to "file_to_search" file in ar file*/
//int ar_get_file(char *ar_filename, char *file_to_search, size_t *file_len);

int abpmArFile::readheader(char *buf) {
	int read_bytes=0;
	int r;
	int len = sizeof(struct ar_hdr);
	
	while(read_bytes<len) {
		r = raw_reader->read(buf + read_bytes, len - read_bytes);
		if(r == 0) {
/*			if(read_bytes != 0)
				fprintf(stderr, "WARNING: EOF in ar header");*/
			//fprintf(stderr,"%s %d couldn't read file. header len = %d\n", __FILE__,__LINE__, read_bytes);
			return 0;
		}
		read_bytes += r;
	}
	return 1;
}

bool abpmArFile::check_magic() {
	int r, read_bytes=0;
	char ar_magic_str[SARMAG];
	while (read_bytes<SARMAG) {
		r = raw_reader->read(ar_magic_str + read_bytes, SARMAG - read_bytes);
		if(r <= 0) {
			return false;
		}
		read_bytes += r;
	}
	
	
	if(strncmp(ar_magic_str, ARMAG, SARMAG))
		return false;
	return true;
}

abpmFile *abpmArFile::choose_file(abpmString file) {
	struct ar_hdr header;
	char size_str[11];
	size_t size;
	int i;
	char c;
	for(i=0; file[i]=='/'; i++);
	abpmString basename=file.tail(i);
	
	raw_reader->seek(0,SEEK_SET);
	
	if(!check_magic()) {
		fprintf(stderr, "Warning!! It's not 'ar' file format!\n");
		return new abpmWrongFile(this, abpmString(__FILE__)+": "+__LINE__+": ");
	}

//	filename[16] = '\0';
	size_str[10] = '\0';
	while(1) {
		if(!readheader((char *)(&header)))
			break;
		memcpy(size_str, header.ar_size, 10);
		//memcpy(filename, header.ar_name, 16);
		for(i=15; i>0; i--) {
			if(header.ar_name[i] == ' ')
				header.ar_name[i] = '\0';
			else
				break;
		}
		size = strtol(size_str, NULL, 0);
		/**/
		if(!strncmp(header.ar_name, basename, 16)) {
			start = raw_reader->seek(0, SEEK_CUR);
			end = start + size;
			file_choosen = true;
			return this;
		}
		raw_reader->seek(size, SEEK_CUR);
		if(raw_reader->read(&c, 1) > 0 && c != '\n')
			raw_reader->seek(-1, SEEK_CUR);
	}
	
	return new abpmWrongFile(this, abpmString(__FILE__)+": "+__LINE__+": ");
}
