#include "ph_types.h"
#include "abpmDebFile.h"
#include "abpmTGZFile.h"
#include "abpmArFile.h"
#include "abpmVector.h"
#include "abpmRawFile.h"
#include "physical_iface.h"
#include "deb_db_parse.h"
#include "item.h"
#include "utils.h"
#include "names.h"
//#include "debug.h"

using namespace physical;
void abpmDebFile::checkcontrol() {
	if(control==NULL) {
		if(data != NULL) {
			delete data;
			data=NULL;
		}
		control = new abpmTGZFile(pkg->choose_file(abpmString("control.tar.gz")), true);
	}
}
void abpmDebFile::checkdata() {
	if(data==NULL) {
		if(control != NULL) {
			delete control;
			control=NULL;
		}
		data = new abpmTGZFile(pkg->choose_file(abpmString("data.tar.gz")), true);
	}
}

abpmString abpmDebFile::file2str(abpmFile *f) {
	char *buf;
	ssize_t read_bytes;
	abpmString result;
	buf = new char[1024];
	while((read_bytes=f->read(buf, 1023))>0) {
		buf[read_bytes]='\0';
		result += abpmString(buf, read_bytes);
		buf=new char[1024];
	}
	delete[] buf;
	return result;
}

abpmBasePackageFile::distfile_type abpmDebFile::cur_distfile_type() {
	abpmString b=base_filename(cur_distfile_name());
	if(conffiles.in(b))
		return CONFIG;
	return NORMAL;
}
bool abpmDebFile::get_physical_info(physical_package_info *info) {
	checkcontrol();
	for(int i=0; i<CNTL_FILE_COUNT; i++) {
		abpmString c_file(control_files[i]);
		if(control->is_file(c_file)) {
			control->choose_file(c_file);
			deb_get_info(i, info, control);
		}
	}
	return true;
}
#include <stdlib.h>
bool abpmDebFile::get_package_info(package_control_info *info) {
	bool retval=true;
	checkcontrol();
	abpmString cntl("control");
	if(!control->is_file(cntl))
		return false;
	control->choose_file(cntl);
	package_control_info *p_info;
	int num;
	read_control_db(control, &p_info, &num);
	if(num != 1) {
		if(num < 1)
			fprintf(stderr, "WARNING: in package %s:\n\terror in reading control file\n", (char *)name);
		else
			fprintf(stderr, "WARNING: in package %s:\n\tmore than one section in control file\n", (char *)name);
		fprintf(stderr, "%s: deb_get_package_info(): %d: num = %d\n",__FILE__,__LINE__, num);
		for(int i=0; i<num; i++)
			free_item(p_info[i].items);
		info->items = NULL;
		retval = false;
	} else
		info->items = p_info[0].items;
	free(p_info);

	return retval;
}
bool abpmDebFile::next_file() {
	checkdata();
	return data->next_file();
}

bool abpmDebFile::copy_file(abpmString path, abpmString prefix_dir) {
	if(data==NULL)
		return false;
	return data->copy_file(path, prefix_dir);
	/*int fd=open(path, O_WRONLY|O_CREATE);
	if(fd<0)
		return false;
	char buf[1024];
	ssize_t recv_bytes;
	while ((recv_bytes=data->read(buf, 1024))>0) {
		ssize_t sent_bytes=0;
		while(sent_bytes<recv_bytes) {
			ssize_t act_sent_bytes = write(fd, buf+sent_bytes, recv_bytes-sent_bytes);
			if(act_sent_bytes<=0) {
				perror("write");
				close(fd);
				return false;
			}
		}
	}
	close(fd);*/
}

abpmString abpmDebFile::cur_distfile_name() {
	if(data==NULL)
		return abpmString("");
	return data->cur_distfile_name();
}
abpmString abpmDebFile::get_preinst_script() {
	checkcontrol();
	abpmString scriptname("preinst");
	if(!control->is_file(scriptname))
		return abpmString("");
	return file2str(control->choose_file(scriptname));
}
abpmString abpmDebFile::get_postinst_script() {
	checkcontrol();
	abpmString scriptname("postinst");
	if(!control->is_file(scriptname))
		return abpmString("");
	return file2str(control->choose_file(scriptname));
}
abpmString abpmDebFile::get_prerm_script() {
	checkcontrol();
	abpmString scriptname("prerm");
	if(!control->is_file(scriptname))
		return abpmString("");
	return file2str(control->choose_file(scriptname));

}
abpmString abpmDebFile::get_postrm_script() {
	checkcontrol();
	abpmString scriptname("postrm");
	if(!control->is_file(scriptname))
		return abpmString("");
	return file2str(control->choose_file(scriptname));
}

#ifndef DEB_INFO_DIR
#define DEB_INFO_DIR "/var/lib/dpkg/info"
#endif
bool abpmDebFile::install_additional_files(abpmString prefix_dir) {
	checkcontrol();
	abpmString infodir = prefix_dir + "/" + DEB_INFO_DIR;
	abpmString infofile_base = infodir + "/" + name + ".";
	while(control->next_file()) {
		abpmString infofile = infofile_base + (char *)(base_filename(control->cur_distfile_name()));
		//fprintf(stderr, "%s: %d: infofile = %s\n", __FILE__,__LINE__, (char *)infofile);
		mkdir_p(infodir, 0755);
		if(!control->copy_file(infofile, prefix_dir));
			//return false;
	}
	checkdata();
	abpmVector<abpmString> files = data->list();
	FILE *f=fopen(infofile_base + "list", "w");
	if(f==NULL) {
		perror(infofile_base + "list");
		return false;
	}
	for(int i=0; i<files.size(); i++)
		fprintf(f, "%s\n", (char *)files[i]);
	fclose(f);
	return true;
}

abpmVector<abpmString> abpmDebFile::list() {
	checkdata();
	return data->list();
}

abpmFile *abpmDebFile::choose_file(abpmString filename) {
	checkdata();
	return data->choose_file(filename);
}

bool abpmDebFile::is_file(abpmString filename) {
	checkdata();
	return data->is_file(filename);
}

abpmVector<abpmString> abpmDebFile::conf_files() {
	checkcontrol();
	abpmVector<abpmString> retval;
	abpmString cf("conffiles");
	if(!control->is_file(cf))
		return retval;
	control->choose_file(cf);
	while(!control->eof()) {
		abpmString line = readline(control);
		if(line.getLength()>0)
			retval.add(line);
	}
	return retval;
}
void abpmDebFile::create() {
	pkg = new abpmArFile(raw_reader);
	control=NULL;
	data=NULL;
	info=NULL;
	checkcontrol();
	get_package_info(&pkg_cntl_info);
	name = abpmString(pkg_cntl_info.items[PKG_NAME]);
	if(control->is_file(abpmString("conffiles"))) {
		//fprintf(stderr, "%s: %d: conffile\n", __FILE__,__LINE__);
		control->choose_file(abpmString("conffiles"));
		while(!control->eof()) {
			abpmString line = base_filename(readline(control));
			//fprintf(stderr, "%s: %d: conffile: %s\n", __FILE__,__LINE__, (char *)line);
			conffiles.add(line);
		}
	}
}
abpmDebFile::abpmDebFile(abpmFile *rawfile, bool useraw) {
	raw_reader = rawfile;
	use_raw = useraw;
	create();
}

abpmDebFile::abpmDebFile(abpmString path, bool useraw) {
	raw_reader = new abpmRawFile(path);
	use_raw = useraw;
	create();
}

abpmDebFile::~abpmDebFile() {
	if(control !=NULL) {
		//fprintf(stderr, "%s:%d: deleting control\n", __FILE__,__LINE__);
		delete control;
	}
	if(data !=NULL) {
		//fprintf(stderr, "%s:%d: deleting data\n", __FILE__,__LINE__);
		delete data;
	}
	if(pkg!=NULL) {
		//fprintf(stderr, "%s:%d: deleting pkg\n", __FILE__,__LINE__);
		delete pkg;
		raw_reader=NULL;
	}
	if(info!=NULL) {
		delete info;
	}
}
