#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include "abpmVirtualFile.h"
#include "abpmGzFile.h"
#include "abpmString.h"
#include "abpmRawFile.h"
#include "abpm_zlib.h"


static off_t abpm_raw_seeker_fun(off_t offset, int whence, void *reader_class_p) {
	abpmGzFile *myclass = (abpmGzFile *)reader_class_p;
	off_t retval = myclass->raw_seek(offset, whence);
	return retval;
}

static ssize_t abpm_raw_reader_fun(void *out_buf, size_t maxlen, void *reader_class_p) {
	abpmGzFile *myclass = (abpmGzFile *)reader_class_p;
	ssize_t retval = myclass->raw_read(out_buf, maxlen);
	return retval;
}

off_t abpmGzFile::raw_seek(off_t offset, int whence) {
	return raw_reader->seek(offset, whence);
}

ssize_t abpmGzFile::raw_read(void *out_buf, size_t maxlen) {
	return raw_reader->read(out_buf, maxlen);
}
off_t abpmGzFile::seek(off_t offset, int whence) {
	if(whence==SEEK_END)
		throw(abpmString(__FILE__)+": "+__LINE__+": abpmGzFile::seek: SEEK_END is not supported.");
	//gzrewind(gzfile);
	off_t retval = gzseek(gzfile, offset, whence);
	return retval;
}

ssize_t abpmGzFile::read(void *out_buf, size_t maxlen) {
	return gzread(gzfile, out_buf, maxlen);
}

abpmGzFile::abpmGzFile(abpmString filename, bool useraw) {
	raw_reader = new abpmVirtualFile(filename);
	abpm_gzFile o;
	o.reader_class = (void *)this;
	o.reader = abpm_raw_reader_fun;
	o.seeker = abpm_raw_seeker_fun;
	gzfile = gzopen(o);
	use_raw = useraw;
}

abpmGzFile::abpmGzFile(abpmFile *rawreader, bool useraw) {
	raw_reader=rawreader;
	abpm_gzFile o;
	o.reader_class = (void *)this;
	o.reader = abpm_raw_reader_fun;
	o.seeker = abpm_raw_seeker_fun;
	gzfile = gzopen(o);
	use_raw = useraw;
}

abpmGzFile::~abpmGzFile() {
	gzclose(gzfile);
}
