/* vi: set ts=4 sw=4:*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
//#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <utime.h>
#include "abpmRpmFile.h"
#include "abpmRawFile.h"
#include "abpmWrongFile.h"
#include "abpmVector.h"
#include "abpmString.h"

#define SIGTAG_SIZE         1000
#define SIGTAG_MD5          1001
#define SIGTAG_PGP          1002

#define RPMTAG_NAME                     1000
#define RPMTAG_VERSION                  1001
#define RPMTAG_RELEASE                  1002
#define RPMTAG_SERIAL                   1003
#define RPMTAG_SUMMARY                  1004
#define RPMTAG_DESCRIPTION              1005
#define RPMTAG_BUILDTIME                1006
#define RPMTAG_BUILDHOST                1007
#define RPMTAG_INSTALLTIME              1008
#define RPMTAG_SIZE                     1009
#define RPMTAG_DISTRIBUTION             1010
#define RPMTAG_VENDOR                   1011
#define RPMTAG_GIF                      1012
#define RPMTAG_XPM                      1013
#define RPMTAG_COPYRIGHT                1014
#define RPMTAG_PACKAGER                 1015
#define RPMTAG_GROUP                    1016
#define RPMTAG_CHANGELOG                1017
#define RPMTAG_SOURCE                   1018
#define RPMTAG_PATCH                    1019
#define RPMTAG_URL                      1020
#define RPMTAG_OS                       1021
#define RPMTAG_ARCH                     1022
#define RPMTAG_PREIN                    1023
#define RPMTAG_POSTIN                   1024
#define RPMTAG_PREUN                    1025
#define RPMTAG_POSTUN                   1026
#define RPMTAG_FILENAMES                1027
#define RPMTAG_FILESIZES                1028
#define RPMTAG_FILESTATES               1029
#define RPMTAG_FILEMODES                1030
#define RPMTAG_FILEUIDS                 1031
#define RPMTAG_FILEGIDS                 1032
#define RPMTAG_FILERDEVS                1033
#define RPMTAG_FILEMTIMES               1034
#define RPMTAG_FILEMD5S                 1035
#define RPMTAG_FILELINKTOS              1036
#define RPMTAG_FILEFLAGS                1037
#define RPMTAG_ROOT                     1038
#define RPMTAG_FILEUSERNAME             1039
#define RPMTAG_FILEGROUPNAME            1040
#define RPMTAG_EXCLUDE                  1041 /* not used */
#define RPMTAG_EXCLUSIVE                1042 /* not used */
#define RPMTAG_ICON                     1043
#define RPMTAG_SOURCERPM                1044
#define RPMTAG_FILEVERIFYFLAGS          1045
#define RPMTAG_ARCHIVESIZE              1046
#define RPMTAG_PROVIDES                 1047
#define RPMTAG_REQUIREFLAGS             1048
#define RPMTAG_REQUIRENAME              1049
#define RPMTAG_REQUIREVERSION           1050
#define RPMTAG_NOSOURCE                 1051
#define RPMTAG_NOPATCH                  1052
#define RPMTAG_CONFLICTFLAGS            1053
#define RPMTAG_CONFLICTNAME             1054
#define RPMTAG_CONFLICTVERSION          1055
#define RPMTAG_DEFAULTPREFIX            1056
#define RPMTAG_BUILDROOT                1057
#define RPMTAG_INSTALLPREFIX            1058
#define RPMTAG_EXCLUDEARCH              1059
#define RPMTAG_EXCLUDEOS                1060
#define RPMTAG_EXCLUSIVEARCH            1061
#define RPMTAG_EXCLUSIVEOS              1062
#define RPMTAG_AUTOREQPROV              1063 /* used internally by build */
#define RPMTAG_RPMVERSION               1064
#define RPMTAG_TRIGGERSCRIPTS           1065
#define RPMTAG_TRIGGERNAME              1066
#define RPMTAG_TRIGGERVERSION           1067
#define RPMTAG_TRIGGERFLAGS             1068
#define RPMTAG_TRIGGERINDEX             1069
#define RPMTAG_VERIFYSCRIPT             1079

struct rpmlead {
	unsigned char magic[4];
	unsigned char major, minor;
	short type;
	short archnum;
	char name[66];
	short osnum;
	short signature_type;
	char reserved[16];
};

struct rpmsign_hd {
	unsigned char magic[3];
	unsigned char version;
	unsigned char reserved[4];
	u_int32_t index_entry_num;
	u_int32_t signature_bytes;
};

struct rpmsign_index {
	u_int32_t tag;
	u_int32_t type;
	u_int32_t offset;
	u_int32_t count;
};


struct rpmhead {
};

//#include "debug.h"

void abpmRpmFile::gotodata() {
	readlead();
	skipsign(); //SIGN
	skipsign(); //HEADER
	header_read=true;
	start = raw_reader->seek(0, SEEK_CUR);
	end=0x7fffffff;
}

bool abpmRpmFile::readlead() {
	return readbytes((char *)_lead, SIGN_START);
}

bool abpmRpmFile::readuint32(u_int32_t *buf) {
	for(int i=0; i<4; i++)
		if(raw_reader->read((char *)buf + 3 - i, 1)<=0)
			return false;
	return true;
}

bool abpmRpmFile::readsignheader(rpmsign_hd *hd) {
	if(!readbytes((char *)(hd->magic), 3))
		return false;
	if(!readbytes((char *)&(hd->version), 1))
		return false;
	if(!readbytes((char *)(hd->reserved), 4))
		return false;
	if(!readuint32(&(hd->index_entry_num)))
		return false;
	if(!readuint32(&(hd->signature_bytes)))
		return false;
	return true;
}

bool abpmRpmFile::skipsign() {
	rpmsign_hd sign_hd;

	if(!readsignheader(&sign_hd))
		return false;
	off_t offset = sign_hd.index_entry_num * 16 + sign_hd.signature_bytes;
	if(raw_reader->seek(offset, SEEK_CUR) < 0)
		return false;
	while(raw_reader->getchar()==0);
	raw_reader->seek(-1, SEEK_CUR);
	return true;
}

abpmVector<abpmString> abpmRpmFile::list() {
	abpmVector<abpmString> result;
	return result;
}

abpmFile *abpmRpmFile::choose_file(abpmString s) {
	return this;
}

bool abpmRpmFile::is_file(abpmString path) {
	return false;
}

ssize_t abpmRpmFile::read(void *out_buf, size_t len) {
	return raw_reader->read(out_buf, len);
}

off_t abpmRpmFile::seek(off_t offset, int whence) {
	if(!header_read)
		return 0;
	off_t absolute_offset = start + offset; //SEEK_SET
	if(whence==SEEK_CUR)
		absolute_offset = raw_reader->seek(0, SEEK_CUR) + offset;
	if(offset==SEEK_END)
		absolute_offset = end + offset;
	if(absolute_offset > end)
		absolute_offset = end;
	if(absolute_offset < start)
		absolute_offset = start;
	off_t o_offset = raw_reader->seek(absolute_offset, SEEK_SET);
	return o_offset - start;
}

bool abpmRpmFile::eof() {
	return raw_reader->eof();
}

#include <stdio.h>
void abpmRpmFile::printresidoom() {
	readlead();
	skipsign();
	skipsign();
	off_t pos = seek(0, SEEK_CUR);
	//_debuginfo("");
	fprintf(stderr, "%d\n", pos);
//	seek(0, SEEK_SET);
//	for(int i=0; i<pos; i++)
//		putchar(raw_reader->getchar());
//	for(int i=0; i<16; i++)
//		putchar(0xff);
	while(!raw_reader->eof()) {
		putchar(raw_reader->getchar());
	}
}

abpmRpmFile::abpmRpmFile(abpmFile *rawfile, bool useraw) {
	_lead=new rpmlead;
	raw_reader=rawfile;
	use_raw = useraw;
	header_read=false;
}

abpmRpmFile::~abpmRpmFile() {
	delete _lead;
}
