/* vi: set ts=4 sw=4:*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <utime.h>
#include "abpmTGZFile.h"
#include "abpmTarFile.h"
#include "abpmGzFile.h"
#include "abpmVector.h"
#include "abpmString.h"



abpmTGZFile::abpmTGZFile(abpmString filename, bool useraw) {
	abpmGzFile *gzfile = new abpmGzFile(filename);
	tarfile = new abpmTarFile(gzfile);
	raw_reader=tarfile;
	use_raw = useraw;
}

abpmTGZFile::abpmTGZFile(abpmFile *rawreader, bool useraw) {
	abpmGzFile *gzfile = new abpmGzFile(rawreader);
	tarfile = new abpmTarFile(gzfile);
	raw_reader=tarfile;
	use_raw = useraw;
}

abpmTGZFile::abpmTGZFile(abpmFile *rawreader, abpmString subfile_path, bool useraw) {
	abpmGzFile *gzfile = new abpmGzFile(rawreader);
	tarfile = new abpmTarFile(gzfile, subfile_path);
	raw_reader=tarfile;
	use_raw = useraw;
}

off_t abpmTGZFile::seek(off_t offset, int whence) {
	return tarfile->seek(offset,whence);
}

abpmVector<abpmString> abpmTGZFile::list() {
	return tarfile->list();
}

abpmFile *abpmTGZFile::choose_file(abpmString s) {
	return tarfile->choose_file(s);
}

bool abpmTGZFile::unpack(abpmString destdir) {
	return tarfile->unpack(destdir);
}

bool abpmTGZFile::is_file(abpmString path) {
	bool retval = tarfile->is_file(path);
	return retval;
}

ssize_t abpmTGZFile::read(void *buf, size_t maxlen) {
	return tarfile->read(buf,maxlen);
}
