/* vi: set ts=4 sw=4:*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
//#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <utime.h>
#include "abpmTarFile.h"
#include "abpmRawFile.h"
#include "abpmWrongFile.h"
#include "abpmVector.h"
#include "abpmString.h"
#include "utils.h"

//#include "debug.h"

abpmString name_of_oldfile(abpmString path) {
	int i=0;
	abpmString oldname=path+".abpm_old.";
	abpmString retval=oldname+i;
	while(exist(retval)) {
		i++;
		retval=oldname+i;
	}
	return retval;
}

struct tar_header {
	char name[100];
	char mode[8];
	char uid[8];
	char gid[8];
	char size[12];
	char mtime[12];
	char chksum[8];
	char typeflag;
	char linkname[100];
	char magic[6];
	char version[2];
	char uname[32];
	char gname[32];
	char devmajor[8];
	char devminor[8];
	char prefix[155];
	char padding[12];
};


abpmTarFile::abpmTarFile(abpmString filename, bool useraw) {
	file_choosen=false;
	use_raw=useraw;
	raw_reader=new abpmVirtualFile(filename);
}

abpmTarFile::abpmTarFile(abpmFile *rawreader, bool useraw) {
	file_choosen=false;
	use_raw=useraw;
	raw_reader=rawreader;
}

abpmTarFile::abpmTarFile(abpmFile *rawreader, abpmString subfile_path, bool useraw) {
	raw_reader=rawreader;
	use_raw=useraw;
	choose_file(subfile_path);
}

bool abpmTarFile::is_file(abpmString s) {
	off_t orig_pos = raw_reader->seek(0, SEEK_CUR);
	raw_reader->seek(0, SEEK_SET);
	
	tar_info hd;
	tar_header raw_hd;
	
	hd.type = CONTTYPE; //we must set a "not REGTYPE" type.
	
	abpmString b = base_filename(s);
	
	while(next_header(&raw_hd, &hd)) {
		if(hd.type == REGTYPE || hd.type == REGTYPE0) {
			if(b.eq(base_filename(hd.name)))
				return true;
			raw_reader->seek(CEIL_TO_BLOCK(hd.size), SEEK_CUR);
		}
	}
	raw_reader->seek(orig_pos, SEEK_SET);
	return false;
}

off_t abpmTarFile::seek(off_t offset, int whence) {
	if(!file_choosen)
		return 0;
	off_t absolute_offset;
	//if(whence==SEEK_SET) //default
	absolute_offset = start + offset;
	if(whence==SEEK_CUR)
		absolute_offset = raw_reader->seek(0, SEEK_CUR) + offset;
	if(offset==SEEK_END)
		absolute_offset = end + offset;
	if(absolute_offset > end)
		absolute_offset = end;
	if(absolute_offset < start)
		absolute_offset = start;
	off_t out_offset = raw_reader->seek(absolute_offset, SEEK_SET);
	return out_offset - start;
}
bool abpmTarFile::eof() {
	if(!file_choosen)
		return true;
	off_t cur_pos = raw_reader->seek(0, SEEK_CUR);
	if(cur_pos >= end || cur_pos < start)
		return true;
	return false;
}
abpmVector<abpmString> abpmTarFile::list() {
	abpmVector<abpmString> contents;
	
	off_t orig_pos = raw_reader->seek(0, SEEK_CUR);
	raw_reader->seek(0, SEEK_SET);
	
	tar_info hd;
	tar_header raw_hd;
	
	hd.type = CONTTYPE; //we must set a "not REGTYPE" type.

	while(next_header(&raw_hd, &hd)) {
		if(hd.type == REGTYPE || hd.type == REGTYPE0) {
			contents.add(hd.name);
			raw_reader->seek(CEIL_TO_BLOCK(hd.size), SEEK_CUR);
		}
	}
	raw_reader->seek(orig_pos, SEEK_SET);
	return contents;
}

bool abpmTarFile::next_header(tar_header *raw_hd, tar_info *hd) {
	tar_block_type block_type;
	do {
		block_type = read_header(raw_hd, hd);
	} while (block_type == TAR_EMPTY_BLOCK);
	//_debuginfo("");
	//fprintf(stderr, "%d, filename = [%s]\n", block_type, (char *)(hd->name));
	return block_type == TAR_NORMAL_BLOCK;
}

abpmString abpmTarFile::cur_distfile_name() {
	return cur_info.name;
}

abpmFile *abpmTarFile::choose_file(abpmString s) {
	bool found = false;
	tar_header raw_hd;
	
	raw_reader->seek(0, SEEK_SET);
	
	cur_info.type = CONTTYPE; //we must set a "not REGTYPE" type.
	
	abpmString b = base_filename(s);
	while(next_header(&raw_hd, &cur_info)) {
		abpmString bb = base_filename(cur_info.name);
		if(b.eq(base_filename(cur_info.name))) {
			//printf("megvan...\n");
			found = true;
			break;
		}
		if(cur_info.type == REGTYPE || cur_info.type == REGTYPE0) {
			raw_reader->seek(CEIL_TO_BLOCK(cur_info.size), SEEK_CUR);
		}
	}
	if(!found) {
		return new abpmWrongFile(this, abpmString(__FILE__)+": "+__LINE__+": ");
	}
	start = raw_reader->seek(0, SEEK_CUR);
	end = start + cur_info.size;
	file_choosen = true;
	return this;
}

tar_block_type abpmTarFile::read_blocks(void *buf, size_t bytes) {
	size_t s = 0;
	int r;
	while(s < bytes) {
		r = raw_reader->read(((char *)buf) + s, bytes - s);
		if(r <= 0) {
			//if(r < 0) {
			//	return TAR_ERROR;
			//}
			return TAR_END;
		}
		s += r;
	}
	return ((char *)buf)[0] == '\0' ? TAR_EMPTY_BLOCK : TAR_NORMAL_BLOCK;
}

tar_block_type abpmTarFile::read_block(void *buf) {
	return read_blocks(buf, TAR_BLOCK_SIZE);
}

tar_block_type abpmTarFile::read_header(tar_header *raw_hd, tar_info *hd) {
	size_t name_block_size;
	tar_block_type retval;

	retval = read_block((void *)raw_hd);
	if(retval <= TAR_EMPTY_BLOCK)
		return retval;

	if(raw_hd->size[11] != '\0') {
		fprintf(stderr, "WARNING: wrong header in tar archive!!!\n");
		fprintf(stderr, " --\tname = %s\n", (char *)(raw_hd->name));
		raw_hd->size[11] = '\0';
		fprintf(stderr, " --\tsize = %s\n", raw_hd->size);
		//for(int i=0; i<TAR_BLOCK_SIZE; i++)
		//	fputc(((char *)raw_hd)[i], stderr);
		//exit(0);
	}
	hd->size = strtol(raw_hd->size, NULL, 8);
	if(raw_hd->typeflag == GNULONGNAME || raw_hd->typeflag == GNULONGLINK) {
		getchar();
		name_block_size = CEIL_TO_BLOCK(hd->size);

		char *longname = new char[name_block_size+1];
		retval = read_blocks(longname, name_block_size);
		longname[hd->size] = '\0';
		hd->name=abpmString(longname, hd->size);
		if(retval <= 0) {
			fprintf(stderr, "WARNING: tar format error\n");
			return retval;
		}
		while((retval = read_block((void *)raw_hd)) == TAR_EMPTY_BLOCK);
		if(retval <= TAR_EMPTY_BLOCK) {
			fprintf(stderr, "WARNING: tar format error\n");
			return retval;
		}
		hd->size = strtol(raw_hd->size, NULL, 8);
	} else {
		hd->name=abpmString(raw_hd->name);
	}

	hd->mode  = strtol(raw_hd->mode, NULL, 8);
	hd->uid   = strtol(raw_hd->uid, NULL, 8);
	hd->gid   = strtol(raw_hd->gid, NULL, 8);
	hd->size  = strtol(raw_hd->size, NULL, 8);
	hd->mtime = strtol(raw_hd->mtime, NULL, 8);
	//chksum = strtol(raw_hd->chksum, NULL, 8);
	hd->type  = (tar_file_type)(raw_hd->typeflag);
	hd->linkname = abpmString(raw_hd->linkname);
	hd->devmajor  = strtol(raw_hd->devmajor, NULL, 8);
	hd->devminor  = strtol(raw_hd->devminor, NULL, 8);
	return TAR_NORMAL_BLOCK;
}


ssize_t abpmTarFile::read(void *buf, size_t maxlen) {
	size_t real_maxlen = maxlen;
	off_t cur_pos = raw_reader->seek(0, SEEK_CUR);
	if(cur_pos >= end)
		return 0;
	if((off_t)(cur_pos + maxlen) > end)
		real_maxlen = end - cur_pos;
	return raw_reader->read(buf, real_maxlen);
}
bool abpmTarFile::next_file() {
	if(file_choosen)
		raw_reader->seek(CEIL_TO_BLOCK(end), SEEK_SET);
	else
		raw_reader->seek(0, SEEK_SET);
	tar_header raw_hd;
	//_debuginfo("");
	//fprintf(stderr, "pos: %d\n", raw_reader->seek(0, SEEK_CUR));
	if(!next_header(&raw_hd, &cur_info))
		return false;
	//fprintf(stderr, "pos: %d\n", raw_reader->seek(0, SEEK_CUR));
	file_choosen = true;
	start = raw_reader->seek(0, SEEK_CUR);
	end = start + cur_info.size;
	//fprintf(stderr, "%s:%d: name = %s size = %d; end=%d\n", __FILE__,__LINE__, (char *)(cur_info.name),cur_info.size, end);
	return true;
}

bool abpmTarFile::copy_file(abpmString path, abpmString prefix_dir) {
	if(!file_choosen)
		return false;
	/*TODO:ide (is) kell garancia, hogy nem irom felul a fajlt*/

	bool retval=false;
			
	switch(cur_info.type) {
		case REGTYPE:
		case REGTYPE0:
			retval = untar_regular_file(&cur_info, path);
			break;
		case SYMTYPE:
			retval = untar_symlink(&cur_info, path);
			break;
		case DIRTYPE:
			retval = untar_dir(&cur_info, path);
			break;
		case LNKTYPE:
			retval = untar_hardlink(&cur_info, path, prefix_dir);
			break;
		case FIFOTYPE:
			retval = untar_fifo(&cur_info, path);
			break;
		default:
			fprintf(stderr, "untar: Sorry, type '%c' is not implemented yet\n", cur_info.type);
			//TODO: a tbbi tpus...
	}
	return retval;
}

bool abpmTarFile::unpack(abpmString destdir) {
	bool retval = true;

	tar_info hd;
	tar_header raw_hd;
	tar_block_type blk;

	while((blk = read_header(&raw_hd, &hd)) > TAR_END)
		if(blk > TAR_EMPTY_BLOCK) {
			/*TODO:ide (is) kell garancia, hogy nem irom felul a fajlt*/

			//if(dummy)
			//	printf("dummy olvasas indul: name = %s, size = %d\n", hd->name, hd->size);
			switch(hd.type) {
				case REGTYPE:
				case REGTYPE0:
					retval = untar_regular_file(&hd, destdir, NULL);
					break;
				case SYMTYPE:
					retval = untar_symlink(&hd, destdir, NULL);
					break;
				case DIRTYPE:
					retval = untar_dir(&hd, destdir, NULL);
					break;
				case LNKTYPE:
					retval = untar_hardlink(&hd, destdir, NULL);
					break;
				case FIFOTYPE:
					retval = untar_fifo(&hd, destdir, NULL);
					break;
				default:
					fprintf(stderr, "untar: Sorry, type '%c' is not implemented yet\n", hd.type);
					//TODO: a tbbi tpus...
			}
		}
	return retval;
}

ssize_t abpmTarFile::write_out(int out_fd, char *buf, size_t count) {
	size_t act_written_bytes;
	size_t written_bytes = 0;
	if(out_fd != -1) {
		while(written_bytes < count) {
			act_written_bytes = write(out_fd, buf + written_bytes, count - written_bytes);
			if(act_written_bytes <= 0) {
				perror("write");
				return written_bytes;
			}
			written_bytes += act_written_bytes;
		}
	} else {
		fprintf(stderr, "Heeejjj!!!! Hova irjam????\n");
		return (ssize_t)-1;
	}
	return count;
}

abpmString abpmTarFile::set_real_filename(tar_info *hd, abpmString prefix_dir) {
	abpmString base = base_filename(hd->name);
	fprintf(stderr, "%s:%d: name=%s, base=%s\n", __FILE__,__LINE__,(char *)(hd->name), (char *)base);
	return prefix_dir + "/" + base;
}

bool abpmTarFile::untar_regular_file(tar_info *hd, abpmString real_filename) {
	bool retval = true;
	int out_fd = -1;
	char buf[TAR_BLOCK_SIZE];
	ssize_t read_bytes = 0;
	ssize_t act_read_bytes;
	ssize_t bytes_to_read;
	ssize_t act_bytes_to_read;
	struct utimbuf t;
	bytes_to_read = CEIL_TO_BLOCK(hd->size);
	
	out_fd = open(real_filename, O_CREAT | O_WRONLY | O_EXCL, hd->mode);

	
	if(out_fd == -1) {
//		if(exist(real_filename)) {
		if(errno == EEXIST) {
			if(rename(real_filename, name_of_oldfile(real_filename))==0)
				out_fd = open(real_filename, O_CREAT | O_WRONLY | O_EXCL, hd->mode);
			else
				perror(real_filename);
		}
	}
	if(out_fd == -1) {
		perror(hd->name + "->" + real_filename);
		char *dir=dirof(real_filename);
		fprintf(stderr, "%s: %d: mode=%o, %d", __FILE__,__LINE__, hd->mode, is_dir(dir));
		delete[] dir;
		return false;
	}
	if(fchown(out_fd, hd->uid, hd->gid) < 0) {
		perror("chown");
	}
	t.actime = time(0);
	t.modtime = hd->mtime;
	while(read_bytes < bytes_to_read) {
		act_bytes_to_read = bytes_to_read - read_bytes > TAR_BLOCK_SIZE ?
			TAR_BLOCK_SIZE :
			bytes_to_read - read_bytes;

		act_read_bytes = raw_reader->read(buf, act_bytes_to_read);
		read_bytes += act_read_bytes;
		if(act_read_bytes < 0) {
			perror("read");
			retval = false;
			break;
		}
		if(act_read_bytes == 0) {
			fprintf(stderr, "WARNING: unexpected End Of File\n");
			retval = false;
			break;
		}

		if(read_bytes > (ssize_t)hd->size)
			act_read_bytes -= read_bytes - hd->size;
		if(act_read_bytes > 0 && write_out(out_fd, buf, act_read_bytes) != act_read_bytes) {
			retval = false;
			break;
		}
	}
	close(out_fd);
	utime(hd->name, &t);
	return retval;
}

bool abpmTarFile::untar_regular_file(tar_info *hd, abpmString prefix_dir, char *packed_filename, DUMMY_FLAG dummy) {
	bool retval = true;

	if(packed_filename && !(hd->name.eq(packed_filename))) {
		retval = false;
		dummy = DUMMY;
	}

	if(dummy == REALLY) {
		abpmString real_filename = set_real_filename(hd, prefix_dir);

		retval=untar_regular_file(hd, real_filename);
	}// else
		//printf("DUMMY READ: size = %d  %d\n", hd->size, bytes_to_read);

// else
	//	printf("DUMMY READ end\n");
	return retval;
}

bool abpmTarFile::untar_dir(tar_info *hd, abpmString real_dir) {
	if(mkdir(real_dir, hd->mode) == 0)
		return true;
	
	if(errno == EEXIST)
		if(is_dir(real_dir))
			return true;
	
	if(rename(real_dir, name_of_oldfile(real_dir))<0) {
		perror(real_dir);
		return false;
	}
	if(mkdir(real_dir, hd->mode) == 0)
		return true;
	perror(real_dir);
	return false;
}

bool abpmTarFile::untar_dir(tar_info *hd, abpmString prefix_dir, char *packed_filename, DUMMY_FLAG dummy) {

	if(packed_filename && !hd->name.eq(packed_filename))
		return false;

	if(dummy == DUMMY)
		return true;

	abpmString real_dir = set_real_filename(hd, prefix_dir);

	//printf("mkdir %s\n", real_dir);
	return untar_dir(hd, real_dir);
}

bool abpmTarFile::untar_symlink(tar_info *hd, abpmString real_name) {
	if(symlink(hd->linkname, real_name) == 0)
		return true;

	//fprintf(stderr, "%s:%d: linkname=%s, real_name=%s\n", __FILE__,__LINE__,(char *)(hd->linkname), (char *)real_name);
	if(errno == EEXIST) {
		if(rename(real_name, name_of_oldfile(real_name))<0) {
			perror(real_name);
			return false;
		}
	} else {
		perror(real_name);
		return false;
	}
	if(symlink(hd->linkname, real_name) == 0)
		return true;
	perror(real_name);
	return false;
}

bool abpmTarFile::untar_symlink(tar_info *hd, abpmString prefix_dir, char *packed_filename, DUMMY_FLAG dummy) {
	if(packed_filename && !hd->name.eq(packed_filename))
		return false;

	if(dummy == DUMMY)
		return true;

	abpmString real_name = set_real_filename(hd, prefix_dir);
	
	return untar_symlink(hd, real_name);
}

bool abpmTarFile::untar_hardlink(tar_info *hd, abpmString real_name, abpmString prefix_dir) {
	abpmString base = base_filename(hd->linkname);
	abpmString realfn = prefix_dir + "/" + base;
	if(link(realfn, real_name) == 0)
		return true;
		//fprintf(stderr, "%s:%d: linkname=%s, base=%s\n", __FILE__,__LINE__,(char *)(hd->linkname), (char *)realfn);
	if(errno == EEXIST) {
		if(rename(real_name, name_of_oldfile(real_name))<0) {
			perror(real_name);
			return false;
		}
	} else {
		perror(real_name);
		return false;
	}
	if(link(realfn, real_name) == 0)
		return true;
	perror(real_name);
	return false;
}


bool abpmTarFile::untar_hardlink(tar_info *hd, abpmString prefix_dir, char *packed_filename, DUMMY_FLAG dummy) {
	if(packed_filename && !hd->name.eq(packed_filename))
		return false;

	if(dummy == DUMMY)
		return true;

	abpmString real_name = set_real_filename(hd, prefix_dir);

	return untar_hardlink(hd, real_name, prefix_dir);
}

bool abpmTarFile::untar_fifo(tar_info *hd, abpmString real_name) {
	if(mknod(real_name, hd->mode, S_IFIFO) == 0)
		return true;
	if(errno == EEXIST) {
		if(rename(real_name, name_of_oldfile(real_name))<0) {
			perror(real_name);
			return false;
		}
	} else {
		perror(real_name);
		return false;
	}
	if(mknod(real_name, hd->mode, S_IFIFO) == 0)
		return true;

	perror(real_name);
	return false;
}

bool abpmTarFile::untar_fifo(tar_info *hd, abpmString prefix_dir, char *packed_filename, DUMMY_FLAG dummy) {
	if(packed_filename && !hd->name.eq(packed_filename))
		return false;
	
	if(dummy == DUMMY)
		return true;

	abpmString real_name = set_real_filename(hd, prefix_dir);
	return untar_fifo(hd,real_name);
}
/*
int untar_file_blocks(file_reader *fr, tar_info *hd, char *prefix_dir, char *packed_filename, DUMMY_FLAG dummy) {
	int retval = 0;

	//if(dummy)
	//	printf("dummy olvasas indul: name = %s, size = %d\n", hd->name, hd->size);
	switch(hd->type) {
		case REGTYPE:
		case REGTYPE0:
			retval = untar_regular_file(fr, hd, prefix_dir, packed_filename, dummy);
			break;
		case SYMTYPE:
			retval = untar_symlink(fr, hd, prefix_dir, packed_filename, dummy);
			break;
		case DIRTYPE:
			retval = untar_dir(fr, hd, prefix_dir, packed_filename, dummy);
			break;
		case LNKTYPE:
			retval = untar_hardlink(fr, hd, prefix_dir, packed_filename, dummy);
			break;
		case FIFOTYPE:
			retval = untar_fifo(fr, hd, prefix_dir, packed_filename, dummy);
			break;
		default:
			fprintf(stderr, "untar: Sorry, type '%c' is not implemented yet\n", hd->type);
			//TODO: a tbbi tpus...
	}

	if(hd->name_must_free) {
		free(hd->name);
		hd->name = NULL;
	}
	return retval;
}


*/
/*int main(int argc, char *argv[]) {
  FILE_DESC fd;
  int raw_fd;
  int retval;
  tar_info hd;
  tar_header raw_hd;

  raw_fd = open(argv[1], O_RDONLY);

  if(raw_fd==-1) {
  perror(argv[1]);
  return 1;
  }
  fd = tar_open(raw_fd, "rb", 12345678);


  while((retval = untar_read_header(fd, &raw_hd, &hd)) > TAR_END)
  if(retval > TAR_EMPTY_BLOCK)
  if(untar_file(fd, &hd) < 0)
  break;

  tar_close(fd);

  return 0;
  }*/
