#ifndef __ABPM_TAR_FILE_H
#define __ABPM_TAR_FILE_H

//#include "abpm_zlib.h"
#include "ph_types.h"
#include "abpmVirtualFile.h"
#include "abpmDirectory.h"
#include <unistd.h>


#define TAR_BLOCK_SIZE 512
#define TAR_BLOCK_SIZE_1 511  /* TAR_BLOCK_SIZE - 1 */

#define CEIL_TO_BLOCK(x) (((x) & TAR_BLOCK_SIZE_1) == 0 ? (x) : ((x)|TAR_BLOCK_SIZE_1) + 1)


enum tar_file_type {
	REGTYPE  = '0',            /* regular file */
	REGTYPE0 = '\0',           /* regular file (ancient bug compat)*/
	LNKTYPE  = '1',            /* hard link */
	SYMTYPE  = '2',            /* symbolic link */
	CHRTYPE  = '3',            /* character special */
	BLKTYPE  = '4',            /* block special */
	DIRTYPE  = '5',            /* directory */
	FIFOTYPE = '6',            /* FIFO special */
	CONTTYPE = '7',            /* reserved */
	GNULONGLINK = 'K',         /* GNU long (>100 chars) link name */
	GNULONGNAME = 'L'         /* GNU long (>100 chars) file name */
};

enum tar_block_type {
	TAR_ERROR = -1,
	TAR_END = 0,
	TAR_EMPTY_BLOCK = 1,
	TAR_NORMAL_BLOCK = 2
};

enum DUMMY_FLAG {
	REALLY = 0,
	DUMMY = 1
};

struct tar_header;

struct tar_info {
	abpmString name;
	mode_t mode;
	uid_t uid;
	gid_t gid;
	size_t size;
	time_t mtime;
	tar_file_type type;
	abpmString linkname;
	long devmajor;
	long devminor;
};


class abpmTarFile : public abpmVirtualFile, public abpmDirectory  {
	private:
		off_t start;
		off_t end;
		bool file_choosen;
		
		tar_info cur_info;
		
		tar_block_type read_blocks(void *buf, size_t len);
		tar_block_type read_block(void *buf);
		tar_block_type read_header(tar_header *raw_hd, tar_info *hd);
		bool next_header(tar_header *raw_hd, tar_info *hd);
		ssize_t write_out(int fd, char *buf, size_t len);
		abpmString set_real_filename(tar_info *hd, abpmString prefix_dir);
		bool untar_regular_file(tar_info *hd, abpmString prefix_dir, char *packed_filename, DUMMY_FLAG dummy=REALLY);
		bool untar_regular_file(tar_info *hd, abpmString path);
		bool untar_dir(tar_info *hd, abpmString prefix_dir, char *packed_filename, DUMMY_FLAG dummy=REALLY);
		bool untar_dir(tar_info *hd, abpmString path);
		bool untar_symlink(tar_info *hd, abpmString prefix_dir, char *packed_filename, DUMMY_FLAG dummy=REALLY);
		bool untar_symlink(tar_info *hd, abpmString path);
		bool untar_hardlink(tar_info *hd, abpmString prefix_dir, char *packed_filename, DUMMY_FLAG dummy=REALLY);
		bool untar_hardlink(tar_info *hd, abpmString path, abpmString prefix_dir);
		bool untar_fifo(tar_info *hd, abpmString prefix_dir, char *packed_filename, DUMMY_FLAG dummy=REALLY);
		bool untar_fifo(tar_info *hd, abpmString path);
		//abpmVector<abpmString> dir_cache;    //TODO...
	public:
		abpmVector<abpmString> list();
		abpmFile *choose_file(abpmString s);
		bool is_file(abpmString path);
		ssize_t read(void *out_buf, size_t maxlen);
		off_t seek(off_t offset, int whence);
		bool eof();
		bool next_file();
		bool copy_file(abpmString path, abpmString prefix_dir);
		abpmString cur_distfile_name();
		abpmTarFile(abpmString filename, bool useraw=false);
		abpmTarFile(abpmFile *raw_reader, bool useraw=false);
		abpmTarFile(abpmFile *raw_reader, abpmString subfile_path, bool useraw=false);

		bool unpack(abpmString destdir);
};

#endif
