#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include "abpmVirtualFile.h"
#include "abpmString.h"
#include "abpmRawFile.h"
#include "abpmGzFile.h"
#include "abpmTarFile.h"
#include "abpmTGZFile.h"
#include "abpmArFile.h"
#include "abpmWrongFile.h"
#include "utils.h"

bool match_end(abpmString s, const char*str) {
	int len=strlen(str);
	if(s.getLength() < len)
		return false;
	for(int i=0; i<len; i++)
		if(str[i]!=s[s.getLength()-len+i])
			return false;
	return true;
}

FILE_TYPES file_type(abpmString file) {
	if(match_end(file,".ar"))
		return AR;
	if(match_end(file,".deb"))
		return AR;
	if(match_end(file,".uhu"))
		return AR;
	if(match_end(file,".tgz"))
		return TGZ;
	if(match_end(file,".tar.gz"))
		return TGZ;
	if(match_end(file,".tar"))
		return TAR;
	if(match_end(file,".gz"))
		return GZ;
	return UNKNOWN;
}


inline ssize_t myread(int fd, void *buf, size_t count) {
	return read(fd, buf, count);
}

bool abpmVirtualFile::readbytes(char *buf, int len) {
	int read_bytes=0;
	int r;
	
	while(read_bytes<len) {
		r = raw_reader->read(buf + read_bytes, len - read_bytes);
		if(r == 0) {
/*			if(read_bytes != 0)
				fprintf(stderr, "WARNING: EOF in ar header");*/
			//fprintf(stderr,"%s %d couldn't read file. header len = %d\n", __FILE__,__LINE__, read_bytes);
			return false;
		}
		read_bytes += r;
	}
	return true;
}
ssize_t abpmVirtualFile::read(void *out_buf, size_t maxlen) {
	return raw_reader->read(out_buf,maxlen);
}

off_t abpmVirtualFile::seek(off_t offset, int whence) {
	return raw_reader->seek(offset, whence);
}

int abpmVirtualFile::getchar() {
	char c;
	if(read(&c, 1)<=0)
		return EOF;
	return (int)c;
}

bool abpmVirtualFile::eof() {
	return raw_reader->eof();
}

abpmFile *create_File_from_string(abpmString logical_path, abpmDirectory *parent=NULL);

abpmVirtualFile::abpmVirtualFile(abpmString filename, bool useraw) {
	raw_reader = create_File_from_string(filename);
	use_raw = useraw;
}

abpmVirtualFile::abpmVirtualFile(abpmFile *rawreader, bool useraw) {
	raw_reader = rawreader;
	use_raw = useraw;
	
}

bool virtual_check_file(abpmString path, abpmDirectory *dir=NULL) {
	if(dir!=NULL)
		return dir->is_file(path);
	return is_file(path);
}

//#include <stdio.h>
//#include <stdlib.h>
int split_dir(abpmString path, abpmDirectory *dir=NULL) {
	int i=path.getLength();
	abpmString subname=path;
	while (!virtual_check_file(subname, dir)) {
//	fprintf(stderr,"%s: %d: subname=%s\n",__FILE__,__LINE__,(char *)subname);
		for(i=subname.getLength()-1; i > 0 && subname[i] != '/'; i--);
		if(i==0)
			break;
		subname=path.head(i);
	}
//	exit(0);
	return i;
}

abpmString abpmVirtualFile::base_filename(abpmString& s) {
	int i=0;
	while(s[i]=='/') i++;
	if(s[i]=='.') i++;
	while(s[i]=='/') i++;
	return s.tail(i);
}

abpmFile *create_File_from_string(abpmString logical_path, abpmDirectory *parent) {
	int i=split_dir(logical_path, parent);
	abpmString logical_dirname = logical_path.head(i);
	abpmString logical_subname = logical_path.tail(i);
	abpmFile *rawfile;
	if(parent==NULL) {
		rawfile = new abpmRawFile(logical_dirname);
	} else
		rawfile = parent->choose_file(logical_dirname);
	if(logical_subname.eq("")) {
		return rawfile;
	}
	abpmDirectory *subdir=NULL;
	abpmTarFile *tarfile=NULL;
	abpmTGZFile *tgzfile=NULL;
	abpmArFile *arfile=NULL;
	switch(file_type(logical_dirname)) {
		case TAR:
			tarfile = new abpmTarFile(rawfile);
			subdir = tarfile;
			break;
		case AR:
			arfile = new abpmArFile(rawfile);
			subdir = arfile;
			break;
		case TGZ:
			tgzfile = new abpmTGZFile(rawfile);
			subdir=tgzfile;
			break;
		default:
			fprintf(stderr, "%s: %d: wrongfile lesz: %s %s\n",__FILE__,__LINE__,(char *)logical_dirname,(char *) logical_subname);
			return new abpmWrongFile(rawfile, abpmString(__FILE__)+": "+__LINE__+": ");
	}
	return create_File_from_string(logical_subname, subdir);
}

abpmVirtualFile::~abpmVirtualFile() {
	if(raw_reader!=NULL && !use_raw)
		delete raw_reader;
}

