/* vi: set ts=4 sw=4: */
#include <sys/types.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include "ph_types.h"
#include "abpmTarFile.h"
#include "physical_iface.h"
#include "abpm_zlib.h"
#include "deb_cntls.h"
#include "utils.h"

#define whitespace(c) (((c) == ' ') || ((c) == '\t') || ((c) == '\n'))

void push_into_package_info(physical_package_info *info, physical_file_info &file_info) {
	int where = info->find(file_info);
	if(where<0) {
		info->insert_file(file_info);
		return;
	}
	physical_file_info *there = info->getp(where);
	if(file_info.uid_used) {
		there->uid_used=1;
		there->owner_id=file_info.owner_id;
	}
	if(file_info.owner!=NULL)
		there->owner=file_info.owner;
		//XXX ha ketszer adjuk meg egymas utan, akkor memory leak
		// -> check there->owner==NULL;
	if(file_info.gid_used) {
		there->gid_used=1;
		there->group_id=file_info.group_id;
	}
	if(file_info.group!=NULL)
		there->group=file_info.group;
		//XXX
	if(file_info.type_used) {
		there->type_used=1;
		there->type=file_info.type;
	}
	if(file_info.mode_used) {
		there->mode_used=1;
		there->mode=file_info.mode;
	}
	if(file_info.major_used) {
		there->major_used=1;
		there->major = file_info.major;
	}
	if(file_info.minor_used) {
		there->minor_used=1;
		there->minor=file_info.minor;
	}
	if(file_info.size_used) {
		there->size_used=1;
		there->size=file_info.size;
	}
	if(file_info.time_used) {
		there->time_used=1;
		there->time=file_info.time;
	}
	if(file_info.md5 !=NULL)
		there->md5=file_info.md5;
	if(file_info.flags_used) {
		there->flags_used=1;
		there->flags=file_info.flags;
	}
}

void push_parentdirs(physical_package_info *info, const char *filename) {
	abpmString dir(filename);
	dir = abpmString(dirof(dir));
	physical_file_info newfile;
	newfile.uid_used = 0;
	newfile.owner = NULL;
	newfile.gid_used = 0;
	newfile.group = NULL;
	newfile.type_used = 1;
	newfile.type='d'; //We know it will be a directory
	newfile.mode_used = 0;
	newfile.major_used = 0;
	newfile.minor_used = 0;
	newfile.size_used = 0;
	newfile.time_used = 0;
	newfile.md5 = NULL;
	newfile.flags_used = 0;

	while(!dir.eq("") && !dir.eq("/")) {
		newfile.filename = dir;
		if(!info->in(newfile))
			info->insert_file(newfile);
		dir = abpmString(dirof(dir));
	}
}

#define BUF_SIZE 512

abpmString cut_link_dest(abpmString &s) {
	int len=s.getLength();
	for(int i=0; i<len-4; i++)
		if(s.tail(i).eq(" -> ", 4)) {
			return s.head(i);
		}
	return s;
}

void deb_get_cntl_info(physical_package_info *info, abpmFile *fr, PHYSICAL_CNTL_INFO_TYPE *entries_to_watch, int entry_num) {
	char buf[513];
	//size_t bytes_to_read;
	//unsigned int act_bytes_to_read;
	size_t processed_bytes = 0;
	size_t read_bytes = 0;
	unsigned int act_read_bytes;
	PHYSICAL_CNTL_INFO_TYPE string_type;
	int act_entry = 0;
	int str_index = 0;
	char *newstr = NULL;
	char *endptr;
	unsigned int i;
	physical_file_info newfile;
	newfile.uid_used = 0;
	newfile.owner = NULL;
	newfile.gid_used = 0;
	newfile.group = NULL;
	newfile.type_used = 0;
	newfile.mode_used = 0;
	newfile.major_used = 0;
	newfile.minor_used = 0;
	newfile.size_used = 0;
	newfile.time_used = 0;
	newfile.md5 = NULL;
	newfile.flags_used = 0;

	string_type = entries_to_watch[act_entry];
	
	//bytes_to_read = CEIL_TO_BLOCK(len);
	
	buf[BUF_SIZE] = '\0';
	
	while(!fr->eof()) {
		act_read_bytes = fr->read(buf, BUF_SIZE);
		read_bytes += act_read_bytes;
		for(i=0; i<act_read_bytes;) {
			if(str_index == 0)
				while(i < act_read_bytes && whitespace(buf[i])) {
					i++;
					processed_bytes++;
				}

			while(i < act_read_bytes && !((whitespace(buf[i]) && string_type != NAME) || buf[i] == '\n')) {
				if((str_index & 0xff) == 0) {
					newstr = (char *)realloc(newstr, (str_index + 0x101) * sizeof(char));
					if(newstr == NULL) {
						perror("realloc");
						exit(1);
					}
				}
				
				newstr[str_index] = buf[i];
				i++;
				str_index++;
			}
			
			if((whitespace(buf[i]) && string_type != NAME) || buf[i] == '\n') {
				newstr[str_index] = '\0';
				processed_bytes += str_index;
									
				switch(string_type) {
					case OWNER_ID:
						newfile.owner_id = strtol(newstr, &endptr, 0);
						if(endptr != newstr)
							newfile.uid_used = 1;
						free(newstr);
						break;
					case OWNER:
						newfile.owner = newstr;
						break;
					case GROUP_ID:
						newfile.group_id = strtol(newstr, &endptr, 0);
						if(endptr != newstr)
							newfile.gid_used = 1;
						free(newstr);
						break;
					case GROUP:
						newfile.group = newstr;
						break;
					case TYPE:
						newfile.type = newstr[0];
						newfile.type_used = 1;
						free(newstr);
						break;
					case MODE:
						newfile.mode = strtol(newstr, &endptr, 8);
						if(endptr != newstr)
							newfile.mode_used = 1;
						free(newstr);
						break;
					case MAJOR:
						newfile.major = strtol(newstr, &endptr, 0);
						if(endptr != newstr) {
							newfile.major_used = 1;
//							printf("major unused\n");
						}
						free(newstr);
						break;
					case MINOR:
						newfile.minor = strtol(newstr, &endptr, 0);
						if(endptr != newstr)
							newfile.minor_used = 1;
						free(newstr);
						break;
					case SIZE:
						newfile.size = strtol(newstr, &endptr, 0);
						if(endptr != newstr)
							newfile.size_used = 1;
						free(newstr);
						break;
					case TIME:
						/*XXX Ez biztos gy van???????????*/
						newfile.time = strtol(newstr, &endptr, 0);
						if(endptr != newstr)
							newfile.time_used = 1;
						free(newstr);
						break;
					case MD5:
						newfile.md5 = newstr;
						break;
					case NAME:
//						printf("newstr=%s\n", newstr);
						if(newstr[0] == '/')
							newfile.filename = abpmString(newstr);
						else {
							newfile.filename = abpmString("/") + newstr;
						}
						free(newstr);
						break;
				}
				newstr = NULL;
				str_index = 0;
				if(buf[i] == '\n') {
//					printf("betesszuk..\n");
					if(newfile.type=='l')
						newfile.filename = cut_link_dest(newfile.filename);
					//fprintf(stderr, "%s:%d: filename=%s\n", __FILE__,__LINE__, newfile.filename);
					push_into_package_info(info, newfile);
					push_parentdirs(info, newfile.filename);
				}
				act_entry = (act_entry + 1) % entry_num;
				string_type = entries_to_watch[act_entry];
//				printf("deb_get_cntl_info: string_type = %d\n", string_type);
			}// else
//				printf("nem whitespace volt, megyonk tovabb... [%s]\n buf[%d] = %c\n", newstr, i, buf[i]);
			while(i < act_read_bytes && whitespace(buf[i])) {
				i++;
				processed_bytes++;
			}
		}
	}
}
