#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>
#include "physical_iface.h"
#include "abpmString.h"
#include "abpmVirtualFile.h"
#include "abpmTGZFile.h"
#include "abpmDebFile.h"
#include "deb_cntls.h"
#include "deb_db_parse.h"
#include "item.h"
#include "utils.h"

namespace physical {
	
	const char *control_files[] = {
		"./buildinfo",
		"./control",
		"./md5sums",
		"./postinst",
		"./postrm",
		"./prerm",
		"./stat"
	};

	void deb_get_stat(physical_package_info *info, abpmFile *fr) {
		PHYSICAL_CNTL_INFO_TYPE types[11];


		types[0] = OWNER_ID;
		types[1] = OWNER;
		types[2] = GROUP_ID;
		types[3] = GROUP;
		types[4] = TYPE;
		types[5] = MODE;
		types[6] = MAJOR;
		types[7] = MINOR;
		types[8] = SIZE;
		types[9] = TIME;
		types[10] = NAME;

		deb_get_cntl_info(info, fr, types, 11);
	}

	void deb_get_md5sums(physical_package_info *info, abpmFile *fr) {
		PHYSICAL_CNTL_INFO_TYPE types[2];

		types[0] = MD5;
		types[1] = NAME;

		deb_get_cntl_info(info, fr, types, 2);
	}


	void deb_get_info(int info_type, physical_package_info *info, abpmFile *fr) {
		switch(info_type) {
			case MD5SUMS:
				deb_get_md5sums(info, fr);
				break;
			case STAT:
				deb_get_stat(info, fr);
				break;
			default:
				//untar_file_blocks(fr, hd, NULL, NULL, DUMMY);
				break;
		}
	}
	//returns 0 if "everything is ok" and -1 if error detected.
	int deb_get_physical_info(char *pkg_filename, physical_package_info *info) {
		abpmString pkg_fn(pkg_filename);
		abpmVirtualFile *pkg = new abpmVirtualFile(pkg_fn+"/control.tar.gz");
		abpmTGZFile *control_reader = new abpmTGZFile(pkg);
		for(int i=0; i<CNTL_FILE_COUNT; i++) {
			abpmString c_file(control_files[i]);
			if(control_reader->is_file(c_file)) {
				control_reader->choose_file(c_file);
				deb_get_info(i, info, control_reader);
			}
		}
		delete control_reader;
		return 0;
	}
	int deb_get_package_info(char *pkg_filename, package_control_info *info) {
		int retval = 0;
		abpmFile *control = new abpmVirtualFile(abpmString(pkg_filename)+"/control.tar.gz/control");
		package_control_info *p_info;
		int num;
		read_control_db(control, &p_info, &num);
		if(num != 1) {
			if(num < 1)
				fprintf(stderr, "WARNING: in package %s:\n\terror in reading control file\n", pkg_filename);
			else
				fprintf(stderr, "WARNING: in package %s:\n\tmore than one section in control file\n", pkg_filename);
			fprintf(stderr, "%s: deb_get_package_info(): %d: num = %d\n",__FILE__,__LINE__, num);
			for(int i=0; i<num; i++)
				free_item(p_info[i].items);
			info->items = NULL;
			retval = -1;
		} else
			info->items = p_info[0].items;

		free(p_info);

		delete control;
		return retval;
	}
	int deb_unpack(char *pkg_filename, char *destdir) {
		abpmDebFile *deb = new abpmDebFile(abpmString(pkg_filename));
		deb->unpack(abpmString(destdir));
		delete deb;
		return 0;
	}
	int deb_copy_scripts(char *pkg_filename, char *destdir) {
		abpmDebFile *deb = new abpmDebFile(abpmString(pkg_filename));
		deb->copy_scripts(abpmString(destdir));
		delete deb;
		return 0;
	}
	int remove(physical_package_info *info, char *prefix_dir) {

		if(prefix_dir == NULL)
			prefix_dir = "";

		for(int i=0; i<info->size(); i++) {
			abpmString filename = (*info)[i].filename;
			abpmString real_filename = abpmString(prefix_dir) + filename;
			//printf("removing file: %s\n", real_filename);
			if((*info)[i].type!='d')
				if(unlink(real_filename))
					perror(real_filename);
			char *parent_dir = dirof(real_filename);
			do {
				//TODO: es ha valamelyik csomag uresen tartalmazza a konyvtarat?
				if(rmdir(parent_dir)!=0) {
					if(errno!=ENOTEMPTY)
						perror(parent_dir);
					break;
				}
				char *new_parent_dir=dirof(parent_dir);
				delete[] parent_dir;
				parent_dir=new_parent_dir;
			} while (parent_dir[0] != '\0' && !strcmp(parent_dir, "/")
					&& !strcmp(parent_dir, prefix_dir));
		}
		return 0;
	}
};

