#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include "abpmFile.h"
#include "utils.h"
#include "abpmString.h"

size_t file_size(char *file) {
	struct stat file_stat;

	if(stat(file, &file_stat) < 0) {
		perror(file);
		return 0;
	}
	return file_stat.st_size;
}

bool exist(char *filename) {
	struct stat file_info;
	if(stat(filename, &file_info)<0) {
		//if(errno != ENOENT)
		//	perror(filename);
		return false;
	}
	return true;
}

bool is_file(char *filename) {
	struct stat file_info;
	if(stat(filename, &file_info)<0) {
		//if(errno != ENOENT)
		//	perror(filename);
		return false;
	}
	if((file_info.st_mode & S_IFDIR) != 0)
		return false;
	
	return true;
}

bool is_dir(char *filename) {
	struct stat file_info;
	if(stat(filename, &file_info)<0) {
		//if(errno != ENOENT)
		//	perror(filename);
		return false;
	}
	if((file_info.st_mode & S_IFDIR) != 0)
		return true;
	
	return false;
}
char *basename(char *file_path) {
	int len = strlen(file_path);
	int i;
	for(i=len - 1; i >= 0 && file_path[i] != '/'; i--);
	
	char *result = file_path + i + 1;
	
	return result;
}


char *dirof(char *file_path) {
	int len = strlen(file_path);
	int i, j, k;
	
	for(i=len - 1; i >= 0 && file_path[i] == '/'; i--);
	for(; i >= 0 && file_path[i] != '/'; i--);
	for(j=i; j>=0 && file_path[j] == '/'; j--);
	j++;
	
	char *result = new char[j+1];
	for(k=0; k<j; k++)
		result[k] = file_path[k];
	result[j] = '\0';
	//printf("dirof: file_path = [%s], result = [%s], j=%d, k=%d, len=%d\n", file_path, result, j, k, len);
	return result;
}


/*mkdir -p*/
bool mkdir_p(char *name, mode_t mode) {
	struct stat file_info;
	
	if(name[0] == '\0')
		return true;
	
	if(stat(name, &file_info)==0 && (file_info.st_mode & S_IFDIR) != 0)
		return true;
	if(errno != ENOENT) {
		perror(name);
		return false;
	}
	
	char *parentdir = dirof(name);
	if(strcmp(parentdir, "") != 0 && strcmp(parentdir, "/") != 0)
		if(!mkdir_p(parentdir, mode)) {
			delete[] parentdir;
			return false;
		}
	delete[] parentdir;
	if(mkdir(name, mode) < 0) {
		perror(name);
		return false;
	}
	return true;
}


abpmString readline(abpmFile *f) {
	abpmString retval;
	int c;
	while((c = f->getchar()) != EOF && c!= '\n')
		retval += (char)c;
	return retval;
}


